from unittest import TestCase
import numpy as np
import pytest

from cr.cube.crunch_cube import (
    CrunchCube,
    _MeanMeasure,
    _Measures,
    _UnweightedCountMeasure,
    _WeightedCountMeasure,
)
from cr.cube.enum import DIMENSION_TYPE as DT
from cr.cube.measures.index import Index
from cr.cube.util import compress_pruned

from ..fixtures import CR  # ---mnemonic: CR = 'cube-response'---
from . import assert_scale_means_equal


class DescribeIntegratedCrunchCube(object):
    def it_provides_a_console_friendly_repr_for_a_cube(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        repr_ = repr(cube)
        assert repr_ == "CrunchCube(name='v4', dim_types='CAT x CAT')"

    def it_provides_access_to_its_dimensions(self, dimensions_fixture):
        cube_response, expected_dimension_types = dimensions_fixture
        cube = CrunchCube(cube_response)

        dimension_types = tuple(d.dimension_type for d in cube.dimensions)

        assert dimension_types == expected_dimension_types

    def it_knows_if_it_is_a_single_ca_cube(self):
        cube = CrunchCube(CR.SIMPLE_CAT_ARRAY)
        is_univariate_ca = cube.is_univariate_ca
        assert is_univariate_ca is True

    def it_knows_the_main_axis_of_a_univariate_ca_cube(self):
        cube = CrunchCube(CR.SIMPLE_CAT_ARRAY)  # ---CA_SUBVAR x CA_CAT---
        univariate_ca_main_axis = cube.univariate_ca_main_axis
        assert univariate_ca_main_axis == 1

    def it_provides_array_for_single_valid_cat_CAT_X_MR(self):
        # --The CAT dim of this cube has only 1 valid category (+1 missing).
        # --This is an edge case because that CAT dimension gets squashed.
        cube = CrunchCube(CR.CAT_X_MR_SENTRY)
        arr = cube.as_array(prune=True)
        np.testing.assert_array_equal(arr, np.array([[0, 0, 0]]))

    def it_provides_pruned_array_for_CA_CAT_x_CA_SUBVAR(self):
        cube = CrunchCube(CR.CA_CAT_X_CA_SUBVAR)
        arr = cube.proportions(axis=0, prune=True)
        np.testing.assert_array_equal(
            compress_pruned(arr),
            np.array(
                [
                    [0.19012797074954296, 0.10494203782794387],
                    [0.2528945764777575, 0.2190359975594875],
                    [0.16514320536258378, 0.1720561317876754],
                    [0.29859841560024375, 0.4069554606467358],
                    [0.09323583180987204, 0.09701037217815742],
                ]
            ),
        )

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            (CR.CA_SUBVAR_HS_X_MR_X_CA_CAT, (DT.CA_SUBVAR, DT.MR, DT.CA_CAT)),
            (CR.CAT_X_LOGICAL, (DT.CAT, DT.LOGICAL)),
            (CR.LOGICAL_UNIVARIATE, (DT.LOGICAL,)),
        ]
    )
    def dimensions_fixture(self, request):
        cube_response, expected_dimension_types = request.param
        return cube_response, expected_dimension_types


class DescribeIntegrated_Measures(object):
    def it_knows_when_its_measures_are_weighted(self, is_weighted_fixture):
        cube_dict, expected_value = is_weighted_fixture
        measures = _Measures(cube_dict, None)

        is_weighted = measures.is_weighted

        assert is_weighted == expected_value

    def it_provides_access_to_the_mean_measure(self):
        cube_dict = CR.CAT_X_CAT_MEAN_WGTD
        measures = _Measures(cube_dict, None)

        means = measures.means

        assert type(means).__name__ == "_MeanMeasure"

    def but_only_when_the_cube_response_contains_means(self):
        cube_dict = CR.CAT_X_CAT
        measures = _Measures(cube_dict, None)

        means = measures.means

        assert means is None

    def it_provides_the_means_missing_count_when_means_are_available(self):
        measures = _Measures(CR.CAT_X_CAT_MEAN_WGTD, None)
        missing_count = measures.missing_count
        assert missing_count == 3

    def but_provides_the_general_missing_count_otherwise(self):
        measures = _Measures(CR.CAT_X_CAT, None)
        missing_count = measures.missing_count
        assert missing_count == 5

    def it_knows_the_population_fraction(self, pop_frac_fixture):
        cube_dict, expected_value = pop_frac_fixture
        measures = _Measures(cube_dict, None)

        population_fraction = measures.population_fraction

        assert population_fraction == expected_value

    def it_provides_access_to_the_unweighted_count_measure(self):
        measures = _Measures(None, None)

        unweighted_counts = measures.unweighted_counts

        assert type(unweighted_counts).__name__ == "_UnweightedCountMeasure"

    def it_knows_the_unweighted_n(self):
        measures = _Measures(CR.CAT_X_CAT, None)
        unweighted_n = measures.unweighted_n
        assert unweighted_n == 20

    def it_provides_access_to_the_weighted_count_measure(self, wgtd_counts_fixture):
        cube_dict, expected_type_name = wgtd_counts_fixture
        measures = _Measures(cube_dict, None)

        weighted_counts = measures.weighted_counts

        assert type(weighted_counts).__name__ == expected_type_name

    def it_knows_the_weighted_n(self, wgtd_n_fixture):
        cube_dict, expected_value = wgtd_n_fixture
        measures = _Measures(cube_dict, None)

        weighted_n = measures.weighted_n

        assert round(weighted_n, 3) == expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            # ---has {'query': {'weight': url}}---
            (CR.ADMIT_X_GENDER_WEIGHTED, True),
            # ---has {'weight_var': weight_name_str}---
            (CR.CAT_X_CAT_X_CAT_WGTD, True),
            # ---unweighted_counts == measure_count_data---
            (CR.ADMIT_X_DEPT_UNWEIGHTED, False),
        ]
    )
    def is_weighted_fixture(self, request):
        cube_response, expected_value = request.param
        cube_dict = cube_response.get("value", cube_response)
        return cube_dict, expected_value

    @pytest.fixture(
        params=[
            # ---filtered case---
            (CR.CAT_X_CAT_FILT, 0.254),
            # ---unfiltered case---
            (CR.CAT_X_CAT, 1.0),
        ]
    )
    def pop_frac_fixture(self, request):
        cube_dict, expected_value = request.param
        return cube_dict, expected_value

    @pytest.fixture(
        params=[
            # ---weighted case---
            (CR.CAT_X_CAT_WGTD, "_WeightedCountMeasure"),
            # ---unweighted case---
            (CR.CAT_X_CAT, "_UnweightedCountMeasure"),
        ]
    )
    def wgtd_counts_fixture(self, request):
        cube_dict, expected_type_name = request.param
        return cube_dict, expected_type_name

    @pytest.fixture(
        params=[
            # ---weighted case---
            (CR.CAT_X_CAT_WGTD, 999.557),
            # ---unweighted case---
            (CR.CAT_X_CAT, 20.0),
        ]
    )
    def wgtd_n_fixture(self, request):
        cube_dict, expected_type = request.param
        return cube_dict, expected_type


class DescribeIntegrated_MeanMeasure(object):
    def it_provides_access_to_its_raw_cube_array(self):
        cube_dict = CR.CAT_X_CAT_MEAN_WGTD
        cube = CrunchCube(cube_dict)
        measure = _MeanMeasure(cube_dict, cube._all_dimensions)

        raw_cube_array = measure.raw_cube_array

        np.testing.assert_array_almost_equal(
            raw_cube_array,
            [
                [52.78205128, 49.90697674, np.nan, np.nan, np.nan],
                [50.43654822, 48.20100503, np.nan, np.nan, np.nan],
                [51.56435644, 47.60283688, np.nan, np.nan, np.nan],
                [58.0, 29.0, np.nan, np.nan, np.nan],
                [37.53846154, 39.45238095, np.nan, np.nan, np.nan],
                [36.66666667, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan],
                [np.nan, np.nan, np.nan, np.nan, np.nan],
            ],
        )


class DescribeIntegrated_UnweightedCountMeasure(object):
    def it_provides_access_to_its_raw_cube_array(self):
        cube_dict = CR.CAT_X_CAT
        cube = CrunchCube(cube_dict)
        measure = _UnweightedCountMeasure(cube_dict, cube._all_dimensions)

        raw_cube_array = measure.raw_cube_array

        np.testing.assert_array_almost_equal(
            raw_cube_array, [[5, 3, 2, 0], [5, 2, 3, 0], [0, 0, 0, 0]]
        )


class DescribeIntegrated_WeightedCountMeasure(object):
    def it_provides_access_to_its_raw_cube_array(self):
        cube_dict = CR.CAT_X_CAT_WGTD
        cube = CrunchCube(cube_dict)
        measure = _WeightedCountMeasure(cube_dict, cube._all_dimensions)

        raw_cube_array = measure.raw_cube_array

        np.testing.assert_array_almost_equal(
            raw_cube_array,
            [
                [32.9, 87.6, 176.2, 117.5, 72.1, 13.4, 0.0, 0.0, 0.0],
                [38.8, 94.1, 199.0128, 102.9, 38.8305, 26.2135, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ],
        )


class TestCrunchCube(TestCase):
    def test_crunch_cube_loads_data(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        cube_dict = cube._cube_dict
        self.assertEqual(cube_dict, CR.CAT_X_CAT)

    def test_as_array_univariate_cat_exclude_missing(self):
        cube = CrunchCube(CR.UNIVARIATE_CATEGORICAL)
        expected = np.array([10, 5])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_numeric(self):
        cube = CrunchCube(CR.VOTER_REGISTRATION)
        expected = np.array([885, 105, 10])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_datetime(self):
        cube = CrunchCube(CR.SIMPLE_DATETIME)
        expected = np.array([1, 1, 1, 1])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_text(self):
        cube = CrunchCube(CR.SIMPLE_TEXT)
        expected = np.array([1, 1, 1, 1, 1, 1])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_cat_x_cat_exclude_missing(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[5, 2], [5, 3]])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_cat_x_cat_unweighted(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[5, 2], [5, 3]])
        actual = cube._as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_cat_x_datetime_exclude_missing(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array(
            [[0, 0, 1, 0], [0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 0]]
        )
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_univariate_cat_axis_none(self):
        cube = CrunchCube(CR.UNIVARIATE_CATEGORICAL)
        expected = np.array([15])
        actual = cube.margin()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_numeric(self):
        cube = CrunchCube(CR.VOTER_REGISTRATION)
        expected = np.array([1000])
        actual = cube.margin()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_datetime(self):
        cube = CrunchCube(CR.SIMPLE_DATETIME)
        expected = np.array([4])
        actual = cube.margin()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_text(self):
        cube = CrunchCube(CR.SIMPLE_TEXT)
        expected = np.array([6])
        actual = cube.margin()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_cat_axis_none(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([15])
        actual = cube.margin()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_datetime_axis_none(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array([4])
        actual = cube.margin()
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_cat_axis_0(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([10, 5])
        actual = cube.margin(axis=0)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_datetime_axis_0(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array([1, 1, 1, 1])
        actual = cube.margin(axis=0)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_cat_axis_1(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([7, 8])
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_datetime_axis_1(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array([1, 1, 1, 1, 0])
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_proportions_univariate_cat_axis_none(self):
        cube = CrunchCube(CR.UNIVARIATE_CATEGORICAL)
        expected = np.array([0.6666667, 0.3333333])
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_numeric(self):
        cube = CrunchCube(CR.VOTER_REGISTRATION)
        expected = np.array([0.885, 0.105, 0.010])
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_datetime(self):
        cube = CrunchCube(CR.SIMPLE_DATETIME)
        expected = np.array([0.25, 0.25, 0.25, 0.25])
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_text(self):
        cube = CrunchCube(CR.SIMPLE_TEXT)
        expected = np.array(
            [0.1666667, 0.1666667, 0.1666667, 0.1666667, 0.1666667, 0.1666667]
        )
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_cat_x_cat_axis_none(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[0.3333333, 0.1333333], [0.3333333, 0.2000000]])
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_cat_x_datetime_axis_none(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array(
            [
                [0.0, 0.0, 0.25, 0.0],
                [0.0, 0.0, 0.0, 0.25],
                [0.0, 0.25, 0.0, 0.0],
                [0.25, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ]
        )
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_cat_x_cat_axis_0(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[0.5, 0.4], [0.5, 0.6]])
        actual = cube.proportions(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_cat_x_datetime_axis_0(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array(
            [[0, 0, 1, 0], [0, 0, 0, 1], [0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 0, 0]]
        )
        actual = cube.proportions(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_cat_x_cat_axis_1(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[0.7142857, 0.2857143], [0.6250000, 0.3750000]])
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_proportions_cat_x_datetime_axis_1(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = np.array(
            [
                [0, 0, 1, 0],
                [0, 0, 0, 1],
                [0, 1, 0, 0],
                [1, 0, 0, 0],
                [np.nan, np.nan, np.nan, np.nan],
            ]
        )
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_univariate_cat_axis_none(self):
        cube = CrunchCube(CR.UNIVARIATE_CATEGORICAL)
        expected = np.array([66.6666667, 33.3333333])
        actual = cube.percentages()
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_numeric(self):
        cube = CrunchCube(CR.VOTER_REGISTRATION)
        expected = np.array([88.5, 10.5, 1.0])
        actual = cube.percentages()
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_datetime(self):
        cube = CrunchCube(CR.SIMPLE_DATETIME)
        expected = np.array([25.0, 25.0, 25.0, 25.0])
        actual = cube.percentages()
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_text(self):
        cube = CrunchCube(CR.SIMPLE_TEXT)
        expected = np.array(
            [16.6666667, 16.6666667, 16.6666667, 16.6666667, 16.6666667, 16.6666667]
        )
        actual = cube.percentages()
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_cat_x_cat_axis_none(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[33.3333333, 13.3333333], [33.3333333, 20.0]])
        actual = cube.percentages()
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_cat_x_cat_axis_0(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[50, 40], [50, 60]])
        actual = cube.percentages(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_percentages_cat_x_cat_axis_1(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[71.4285714, 28.5714286], [62.50000, 37.50000]])
        actual = cube.percentages(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_population_counts_univariate_cat(self):
        cube = CrunchCube(CR.UNIVARIATE_CATEGORICAL)
        expected = np.array([6000.6666667, 3000.3333333])
        actual = cube.population_counts(9001)
        np.testing.assert_almost_equal(actual, expected)

    def test_population_counts_numeric(self):
        cube = CrunchCube(CR.VOTER_REGISTRATION)
        expected = np.array([7965.885, 945.105, 90.01])
        actual = cube.population_counts(9001)
        np.testing.assert_almost_equal(actual, expected)

    def test_population_counts_datetime(self):
        cube = CrunchCube(CR.SIMPLE_DATETIME)
        expected = np.array([2250.25, 2250.25, 2250.25, 2250.25])
        actual = cube.population_counts(9001)
        np.testing.assert_almost_equal(actual, expected)

    def test_population_counts_text(self):
        cube = CrunchCube(CR.SIMPLE_TEXT)
        expected = np.array(
            [
                1500.1666667,
                1500.1666667,
                1500.1666667,
                1500.1666667,
                1500.1666667,
                1500.1666667,
            ]
        )
        actual = cube.population_counts(9001)
        np.testing.assert_almost_equal(actual, expected)

    def test_population_counts_cat_x_cat(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = np.array([[3000.3333333, 1200.1333333], [3000.3333333, 1800.2]])
        actual = cube.population_counts(9001)
        np.testing.assert_almost_equal(actual, expected)

    def test_filtered_population_counts(self):
        cube = CrunchCube(CR.CAT_X_CAT_FILT)
        expected = np.array(
            [
                [300000.0, 1400000.0, 0.0, 0.0, 0.0, 0.0],
                [5900000.0, 13200000.0, 0.0, 0.0, 0.0, 0.0],
                [600000.0, 2900000.0, 0.0, 0.0, 0.0, 0.0],
                [100000.0, 100000.0, 0.0, 0.0, 0.0, 0.0],
                [300000.0, 600000.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        actual = cube.population_counts(100000000)
        np.testing.assert_almost_equal(actual, expected)

    def test_labels_cat_x_cat_exclude_missing(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = [["B", "C"], ["C", "E"]]
        actual = cube.labels()
        self.assertEqual(actual, expected)

    def test_labels_cat_x_datetime_exclude_missing(self):
        cube = CrunchCube(CR.CAT_X_DATETIME)
        expected = [
            ["red", "green", "blue", "4", "9"],
            [
                "1776-07-04T00:00:00",
                "1950-12-24T00:00:00",
                "2000-01-01T00:00:00",
                "2000-01-02T00:00:00",
            ],
        ]
        actual = cube.labels()
        self.assertEqual(actual, expected)

    def test_labels_simple_cat_array_exclude_missing(self):
        cube = CrunchCube(CR.SIMPLE_CAT_ARRAY)
        expected = [["ca_subvar_1", "ca_subvar_2", "ca_subvar_3"], ["a", "b", "c", "d"]]
        actual = cube.labels()
        self.assertEqual(actual, expected)

    def test_as_array_simple_cat_array_exclude_missing(self):
        cube = CrunchCube(CR.SIMPLE_CAT_ARRAY)
        expected = np.array([[3, 3, 0, 0], [1, 3, 2, 0], [0, 2, 1, 3]])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_cat_x_num_x_datetime(self):
        """Test 3D cube, slicing accross first (numerical) variable."""
        cube = CrunchCube(CR.CAT_X_NUM_X_DATETIME)
        expected = np.array(
            [
                [[1, 1], [0, 0], [0, 0], [0, 0]],
                [[2, 1], [1, 1], [0, 0], [0, 0]],
                [[0, 0], [2, 3], [0, 0], [0, 0]],
                [[0, 0], [0, 0], [3, 2], [0, 0]],
                [[0, 0], [0, 0], [1, 1], [0, 1]],
            ]
        )
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_proportions_cat_x_num_datetime(self):
        cube = CrunchCube(CR.CAT_X_NUM_X_DATETIME)
        expected = np.array(
            [
                [[0.5, 0.5], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                [[0.4, 0.2], [0.2, 0.2], [0.0, 0.0], [0.0, 0.0]],
                [[0.0, 0.0], [0.4, 0.6], [0.0, 0.0], [0.0, 0.0]],
                [[0.0, 0.0], [0.0, 0.0], [0.6, 0.4], [0.0, 0.0]],
                [[0.0, 0.0], [0.0, 0.0], [0.33333333, 0.33333333], [0.0, 0.33333333]],
            ]
        )
        # Set axis to tuple (1, 2), since we want to do a total for each slice
        # of the 3D cube. This is consistent with how the np.sum works
        # (axis parameter), which is used internally in
        # 'proportions' calculation.
        actual = cube.proportions((1, 2))
        np.testing.assert_almost_equal(actual, expected)

    def test_cat_x_num_x_datetime_margin_by_table(self):
        cube = CrunchCube(CR.CAT_X_NUM_X_DATETIME)
        # Expect total for each slice (of which there are 5) rather than
        # expecting the total for the entire cube.
        expected = np.array([2, 5, 5, 5, 3])
        actual = cube.margin(axis=None)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_num_x_datetime_axis_0(self):
        cube = CrunchCube(CR.CAT_X_NUM_X_DATETIME)
        # Expect resulting margin to have an additional dimension
        # (tabs CAT, with only a single element)
        expected = np.array([[[3, 2], [3, 4], [4, 3], [0, 1]]])
        actual = cube.margin(axis=0)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_num_x_datetime_axis_1(self):
        cube = CrunchCube(CR.CAT_X_NUM_X_DATETIME)
        expected = np.array([[1, 1], [3, 2], [2, 3], [3, 2], [1, 2]])
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_cat_x_num_x_datetime_axis_2(self):
        cube = CrunchCube(CR.CAT_X_NUM_X_DATETIME)
        expected = np.array(
            [[2, 0, 0, 0], [3, 2, 0, 0], [0, 5, 0, 0], [0, 0, 5, 0], [0, 0, 2, 1]]
        )
        actual = cube.margin(axis=2)
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_unweighted_gender_x_ideology(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array([[32, 85, 171, 114, 70, 13], [40, 97, 205, 106, 40, 27]])
        actual = cube.as_array(weighted=False)
        np.testing.assert_array_equal(actual, expected)

    def test_as_array_weighted_gender_x_ideology(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                [
                    32.98969072,
                    87.62886598,
                    176.28865979,
                    117.5257732,
                    72.16494845,
                    13.40206186,
                ],
                [
                    38.83495146,
                    94.17475728,
                    199.02912621,
                    102.91262136,
                    38.83495146,
                    26.21359223,
                ],
            ]
        )
        actual = cube.as_array()
        np.testing.assert_almost_equal(actual, expected)

    def test_margin_weighted_gender_x_ideology_axis_0(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                71.82464218,
                181.80362326,
                375.31778601,
                220.43839456,
                110.99989991,
                39.61565409,
            ]
        )
        actual = cube.margin(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_margin_unweighted_gender_x_ideology_axis_0(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array([72, 182, 376, 220, 110, 40])
        actual = cube.margin(axis=0, weighted=False)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_unweighted_gender_x_ideology_axis_1(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array([485, 515])
        actual = cube.margin(axis=1, weighted=False)
        np.testing.assert_array_equal(actual, expected)

    def test_margin_weighted_gender_x_ideology_axis_1(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array([500, 500])
        actual = cube.margin(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_calculate_standard_error_axis_0(self):
        """Calculate standard error across columns."""
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                [
                    -0.715899626017458,
                    -0.536708837208206,
                    -1.485149675785275,
                    1.114743776586886,
                    3.355236023985286,
                    -2.077040949965248,
                ],
                [
                    0.715899626017462,
                    0.536708837208211,
                    1.485149675785279,
                    -1.114743776586884,
                    -3.355236023985284,
                    2.07704094996525,
                ],
            ]
        )
        actual = cube.zscore()
        np.testing.assert_almost_equal(actual, expected)

    def test_pvals(self):
        cube = CrunchCube(CR.CAT_X_CAT_GERMAN_WEIGHTED)
        expected = np.array(
            [
                [
                    0.1427612835278633,
                    0.0002121209149277,
                    0.0001314694879104,
                    0.1241771485531613,
                    0.0097454103240531,
                    0.0000000000000699,
                    0.8033849539190183,
                    0.4271118723152929,
                ],
                [
                    0.1427612835278667,
                    0.0002121209149277,
                    0.0001314694879104,
                    0.1241771485531642,
                    0.0097454103240533,
                    0.0000000000000699,
                    0.8033849539190183,
                    0.4271118723152956,
                ],
            ]
        )
        # Test without pruning
        actual = cube.pvals()
        np.testing.assert_almost_equal(actual, expected)

        # Test with pruning
        actual = cube.pvals(prune=True)
        np.testing.assert_almost_equal(actual, expected)

        # Test with pruning and H&S
        actual = cube.pvals(prune=True, hs_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_pvals_stats(self):
        cube = CrunchCube(CR.STATS_TEST)
        expected = np.array(
            [
                [
                    0.0436818197570077,
                    0.0002697141695955,
                    0.0913940671748992,
                    0.6836420776424197,
                    0.4681291494279529,
                    0.0000013632752629,
                ],
                [
                    0.0436818197570077,
                    0.0002697141695955,
                    0.0913940671748992,
                    0.6836420776424197,
                    0.4681291494279529,
                    0.0000013632752629,
                ],
            ]
        )
        # Test without pruning
        actual = cube.pvals()
        np.testing.assert_almost_equal(actual, expected)

        # Test with pruning
        actual = cube.pvals(prune=True)
        np.testing.assert_almost_equal(actual, expected)

        # Test with pruning and H&S
        actual = cube.pvals(prune=True, hs_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_mean_age_for_blame_x_gender(self):
        cube = CrunchCube(CR.ECON_MEAN_AGE_BLAME_X_GENDER)
        expected = np.array(
            [
                [52.78205128205122, 49.9069767441861],
                [50.43654822335009, 48.20100502512572],
                [51.5643564356436, 47.602836879432715],
                [58, 29],
                [37.53846153846155, 39.45238095238095],
            ]
        )
        actual = cube.as_array()
        np.testing.assert_almost_equal(actual, expected)

    def test_mean_no_dims(self):
        cube = CrunchCube(CR.ECON_MEAN_NO_DIMS)
        expected = np.array([49.095])
        actual = cube.as_array()
        np.testing.assert_almost_equal(actual, expected)

    def test_z_scores_admit_by_dept_unweighted_rows(self):
        """see
        https://github.com/Crunch-io/whaam/blob/master/base/stats/tests/zvalues-spec.js#L42
        """
        cube = CrunchCube(CR.ADMIT_X_DEPT_UNWEIGHTED)
        expected = np.array(
            [
                [
                    18.04029230689576,
                    13.018394979149804,
                    -2.576739836814013,
                    -3.059526328377408,
                    -7.230244530709987,
                    -19.321410263144653,
                ],
                [
                    -18.040292306895765,
                    -13.018394979149804,
                    2.576739836814013,
                    3.059526328377408,
                    7.230244530709987,
                    19.321410263144653,
                ],
            ]
        )
        actual = cube.zscore()
        np.testing.assert_almost_equal(actual, expected)

    def test_z_scores_admit_by_gender_weighted_rows(self):
        """ see
        https://github.com/Crunch-io/whaam/blob/master/base/stats/tests/zvalues-spec.js#L67
        """
        cube = CrunchCube(CR.ADMIT_X_GENDER_WEIGHTED)
        expected = np.array(
            [
                [9.42561984520692, -9.425619845206922],
                [-9.425619845206922, 9.42561984520692],
            ]
        )
        actual = cube.zscore()
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_dim_names(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = ["Statements agreed with about Climate", "Gender"]
        actual = [dim.name for dim in cube.dimensions]
        self.assertEqual(actual, expected)

    def test_selected_crosstab_as_array(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                [9928.20954289002, 11524.821237084192],
                [9588.843313998908, 9801.254016136965],
                [11697.435357575358, 13095.670425525452],
                [9782.8995547749, 10531.918128023966],
                [4417.596222134318, 3448.380316269752],
                [6179.175512581436, 6490.427474934746],
            ]
        )
        actual = cube.as_array()
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_margin_by_rows(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                21453.03077997421,
                19390.097330135875,
                24793.105783100807,
                20314.817682798865,
                7865.976538404069,
                12669.602987516182,
            ]
        )
        actual = cube.margin(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_margin_by_cols(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                [14566.261567907562, 15607.301233922663],
                [14456.513325488017, 15450.609903833058],
                [14415.136475733132, 15405.898678070093],
                [11485.661204663904, 11912.588886491172],
                [11664.69933815247, 12110.196347286023],
                [11547.413553551738, 11961.575582997419],
            ]
        )
        actual = cube.margin(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_margin_total(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                30173.5628018302,
                29907.1232293211,
                29821.0351538032,
                23398.2500911551,
                23774.8956854385,
                23508.9891365492,
            ]
        )
        actual = cube.margin()
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_proportions_by_rows(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                [0.4627882020361299, 0.5372117979638701],
                [0.4945227014975337, 0.5054772985024663],
                [0.47180193800279874, 0.5281980619972013],
                [0.481564723224583, 0.5184352767754171],
                [0.5616081106479636, 0.4383918893520365],
                [0.48771658580541166, 0.5122834141945883],
            ]
        )
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_proportions_by_cols(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                [0.6815894041587091, 0.7384249886863752],
                [0.6632887957217867, 0.6343603312193796],
                [0.8114689290154947, 0.8500426167391849],
                [0.8517489224566737, 0.8840998567462627],
                [0.3787149667617584, 0.28475015741941767],
                [0.535113381358101, 0.5426064007955989],
            ]
        )
        actual = cube.proportions(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_proportions_by_cell(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = np.array(
            [
                [0.329036700375595, 0.381950958618156],
                [0.320620717695708, 0.327723062528721],
                [0.392254504152701, 0.439142047148397],
                [0.418103897371069, 0.450115632023491],
                [0.185809278853744, 0.14504292098248],
                [0.262843097025161, 0.27608279697761],
            ]
        )
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_selected_crosstab_is_double_mr(self):
        cube = CrunchCube(CR.SELECTED_CROSSTAB_4)
        expected = False
        actual = cube.is_double_mr
        self.assertEqual(actual, expected)

    def test_pets_x_pets_is_double_mr(self):
        cube = CrunchCube(CR.PETS_X_PETS)
        is_double_mr = cube.is_double_mr
        self.assertEqual(is_double_mr, True)

    def test_pets_x_pets_as_array(self):
        cube = CrunchCube(CR.PETS_X_PETS)
        expected = np.array([[40, 14, 18], [14, 34, 16], [18, 16, 38]])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_pets_x_pets_proportions_by_cell(self):
        cube = CrunchCube(CR.PETS_X_PETS)
        expected = np.array(
            [
                [0.5, 0.2, 0.2571429],
                [0.2, 0.4303797, 0.2285714],
                [0.2571429, 0.2285714, 0.5428571],
            ]
        )
        actual = cube.proportions()
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_pets_proportions_by_col(self):
        cube = CrunchCube(CR.PETS_X_PETS)
        expected = np.array(
            [
                [1.0, 0.4827586, 0.4736842],
                [0.4117647, 1.0, 0.4210526],
                [0.5294118, 0.5517241, 1.0],
            ]
        )
        actual = cube.proportions(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_pets_proportions_by_row(self):
        cube = CrunchCube(CR.PETS_X_PETS)
        expected = np.array(
            [
                [1.0, 0.4117647, 0.5294118],
                [0.4827586, 1.0, 0.5517241],
                [0.4736842, 0.4210526, 1.0],
            ]
        )
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_fruit_as_array(self):
        cube = CrunchCube(CR.PETS_X_FRUIT)
        expected = np.array([[12, 28], [12, 22], [12, 26]])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_pets_x_fruit_margin_row(self):
        cube = CrunchCube(CR.PETS_X_FRUIT)
        expected = np.array([40, 34, 38])
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_pets_array_as_array(self):
        cube = CrunchCube(CR.PETS_ARRAY)
        expected = np.array([[45, 34], [40, 40], [32, 38]])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

    def test_pets_array_proportions(self):
        cube = CrunchCube(CR.PETS_ARRAY)
        expected = np.array(
            [[0.5696203, 0.4303797], [0.5000000, 0.500000], [0.4571429, 0.5428571]]
        )
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_array_proportions_bad_directions(self):
        cube = CrunchCube(CR.PETS_ARRAY)
        with self.assertRaises(ValueError):
            cube.proportions(axis=0)

        # This is not bad direction, because the 'None' always
        # figures out what it needs to do
        # with self.assertRaises(ValueError):
        #     cube.proportions(axis=None)

    def test_pets_array_margin_total(self):
        cube = CrunchCube(CR.PETS_ARRAY)
        with self.assertRaises(ValueError):
            cube.margin()

    def test_pets_array_margin_by_row(self):
        cube = CrunchCube(CR.PETS_ARRAY)
        expected = np.array([79, 80, 70])
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_pets_array_margin_by_col_not_allowed_across_items(self):
        """Colum direction is not allowed across items dimension."""
        cube = CrunchCube(CR.PETS_ARRAY)
        with self.assertRaises(ValueError):
            cube.margin(axis=0)

    def test_count_unweighted(self):
        cube = CrunchCube(CR.ADMIT_X_GENDER_WEIGHTED)
        expected = 4526
        actual = cube.count(weighted=False)
        self.assertEqual(actual, expected)

    def test_count_weighted(self):
        cube = CrunchCube(CR.ADMIT_X_GENDER_WEIGHTED)
        expected = 4451.955438803248
        actual = cube.count(weighted=True)
        self.assertEqual(actual, expected)

    def test_econ_x_ideology_index_by_col(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                [0.91861761, 0.96399471, 0.9394101, 1.06629132, 1.30027051, 0.67660435],
                [1.08138239, 1.03600529, 1.0605899, 0.93370868, 0.69972949, 1.32339565],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_econ_x_ideology_index_by_row(self):
        cube = CrunchCube(CR.ECON_GENDER_X_IDEOLOGY_WEIGHTED)
        expected = np.array(
            [
                [0.91861761, 0.96399471, 0.9394101, 1.06629132, 1.30027051, 0.67660435],
                [1.08138239, 1.03600529, 1.0605899, 0.93370868, 0.69972949, 1.32339565],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_fruit_x_pets_proportions_by_cell(self):
        cube = CrunchCube(CR.FRUIT_X_PETS)
        expected = np.array(
            [[0.15, 0.15189873, 0.17142857], [0.35, 0.27848101, 0.37142857]]
        )
        actual = cube.proportions(axis=None)
        np.testing.assert_almost_equal(actual, expected)

    def test_fruit_x_pets_proportions_by_row(self):
        cube = CrunchCube(CR.FRUIT_X_PETS)
        expected = np.array(
            [[0.4285714, 0.48, 0.5217391], [0.5384615, 0.4074074, 0.5531915]]
        )
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_fruit_x_pets_proportions_by_col(self):
        cube = CrunchCube(CR.FRUIT_X_PETS)
        expected = np.array([[0.3, 0.3529412, 0.3157895], [0.7, 0.6470588, 0.6842105]])
        actual = cube.proportions(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_fruit_proportions_by_cell(self):
        cube = CrunchCube(CR.PETS_X_FRUIT)
        expected = np.array(
            [[0.15, 0.35], [0.15189873, 0.27848101], [0.17142857, 0.37142857]]
        )
        actual = cube.proportions(axis=None)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_fruit_proportions_by_col(self):
        cube = CrunchCube(CR.PETS_X_FRUIT)
        expected = np.array(
            [[0.4285714, 0.5384615], [0.48, 0.4074074], [0.5217391, 0.5531915]]
        )
        actual = cube.proportions(axis=0)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_fruit_proportions_by_row(self):
        cube = CrunchCube(CR.PETS_X_FRUIT)
        expected = np.array(
            [[0.3, 0.7], [0.3529412, 0.6470588], [0.3157895, 0.6842105]]
        )
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_cat_x_cat_array_proportions_by_row(self):
        """Get the proportions for each slice of the 3D cube.

        The axis is equal to 2, since this is the dimensions across which
        we have to calculate the margin.
        """
        cube = CrunchCube(CR.FRUIT_X_PETS_ARRAY)
        expected = [
            [[0.52, 0.48], [0.57142857, 0.42857143], [0.47826087, 0.52173913]],
            [
                [0.59259259, 0.40740741],
                [0.46153846, 0.53846154],
                [0.44680851, 0.55319149],
            ],
        ]
        actual = cube.proportions(axis=2)
        np.testing.assert_almost_equal(actual, expected)

    def test_identity_x_period_axis_out_of_bounds(self):
        cube = CrunchCube(CR.NUM_X_NUM_EMPTY)
        # There are margins that have 0 value in this cube. In whaam, they're
        # pruned, so they're not shown. CrunchCube is not responsible for
        # pruning (cr.exporter is).
        expected = np.array([94, 0, 248, 210, 102, 0, 0, 0, 286, 60])
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_ca_with_single_cat(self):
        cube = CrunchCube(CR.CA_SINGLE_CAT)
        # The last 0 of the expectation is not visible in whaam because of
        # pruning, which is not the responsibility of cr.cube.
        expected = np.array([79, 80, 70, 0])
        actual = cube.margin(axis=1, weighted=False)
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_array_x_pets_by_col(self):
        cube = CrunchCube(CR.PETS_ARRAY_X_PETS)
        expected = np.array(
            [[0.59097127, 0.0, 0.55956679], [0.40902873, 1.0, 0.44043321]]
        )
        # Since cube is 3D, col dim is 1 (instead of 0)
        actual = cube.proportions(axis=1)[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_array_x_pets_row(self):
        cube = CrunchCube(CR.PETS_ARRAY_X_PETS)
        expected = np.array(
            [[0.44836533, 0.0, 0.48261546], [0.39084967, 1.0, 0.47843137]]
        )
        # Since cube is 3D, row dim is 2 (instead of 0)
        actual = cube.proportions(axis=2)[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_array_x_pets_cell(self):
        cube = CrunchCube(CR.PETS_ARRAY_X_PETS)
        expected = np.array(
            [[0.24992768, 0.00000000, 0.26901938], [0.17298235, 0.44258027, 0.21174429]]
        )
        actual = cube.proportions(axis=None)[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_pets_x_pets_array_margin_by_cell(self):
        cube = CrunchCube(CR.PETS_X_PETS_ARRAY)
        # TODO: Confirm with Jon and Mike
        # This margin is not something a user would expect, because it
        # performs summation across the items dimension of the CA.
        # expected = np.array([229, 229, 229])
        with self.assertRaises(ValueError):
            cube.margin()

    def test_pets_x_pets_array_percentages(self):
        """All directions need to return same percentages.

        The only direction that makes sense is across categories, and this is
        handled automatically by the cube.
        """
        cube = CrunchCube(CR.PETS_X_PETS_ARRAY)
        expected = np.array(
            [
                [0.58823529, 0.41176471],
                [0.00000000, 1.00000000],
                [0.47058824, 0.52941176],
            ]
        )
        # TODO: Remove this if not needed anymore...
        # TODO: Consult with jon and Mike. The new expectation is closer to what
        # whaam displays, but diverges from R.
        # expected = np.array([
        #     [0.55555556, 0.19444444],
        #     [0., 0.55555556],
        #     [0.44444444, 0.25],
        # ])

        # The direction 1 designates columns (for each slice). Since this is
        # the subvar dimension, it needs to be treated as invalid direction
        with self.assertRaises(ValueError):
            cube.proportions(axis=1)

        # The direction None designates "table" (total for each slice). Since
        # this would contain the subvar dimension, it needs to be treated as
        # invalid direction
        with self.assertRaises(ValueError):
            cube.proportions()[0]

        actual = cube.proportions(axis=2)[0]
        np.testing.assert_almost_equal(actual, expected)

    def test_profiles_percentages_add_up_to_100(self):
        cube = CrunchCube(CR.PROFILES_PERCENTS)
        props = cube.percentages(axis=1)
        actual_sum = np.sum(props, axis=1)
        expected_sum = np.ones(props.shape[0]) * 100
        np.testing.assert_almost_equal(actual_sum, expected_sum)

    def test_cat_x_cat_as_array_prune_cols(self):
        cube = CrunchCube(CR.CAT_X_CAT_WITH_EMPTY_COLS)
        expected = np.array(
            [
                [2, 2, 0, 1],
                [0, 0, 0, 0],
                [0, 1, 0, 2],
                [0, 2, 0, 0],
                [0, 2, 0, 1],
                [0, 1, 0, 0],
            ]
        )
        actual = cube.as_array(prune=False)
        expected = np.array([[2, 2, 1], [0, 1, 2], [0, 2, 0], [0, 2, 1], [0, 1, 0]])
        table = cube.as_array(prune=True)
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_array_equal(actual, expected)

        pruned_expected = [
            np.array([False, True, False, False, False, False]),
            np.array([False, False, True, False]),
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(pruned[i], pruned_expected[i])

    def test_cat_x_cat_props_by_col_prune_cols(self):
        cube = CrunchCube(CR.CAT_X_CAT_WITH_EMPTY_COLS)
        expected = np.array(
            [
                [1.0, 0.25, np.nan, 0.25],
                [0.0, 0.0, np.nan, 0.0],
                [0.0, 0.125, np.nan, 0.5],
                [0.0, 0.25, np.nan, 0.0],
                [0.0, 0.25, np.nan, 0.25],
                [0.0, 0.125, np.nan, 0.0],
            ]
        )
        actual = cube.proportions(axis=0, prune=False)
        expected = np.array(
            [
                [1.0, 0.25, 0.25],
                [0.0, 0.125, 0.5],
                [0.0, 0.25, 0.0],
                [0.0, 0.25, 0.25],
                [0.0, 0.125, 0.0],
            ]
        )
        table = cube.proportions(axis=0, prune=True)
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_array_equal(actual, expected)

        pruned_expected = [
            np.array([False, True, False, False, False, False]),
            np.array([False, False, True, False]),
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(pruned[i], pruned_expected[i])

    def test_cat_x_cat_props_by_row_prune_cols(self):
        cube = CrunchCube(CR.CAT_X_CAT_WITH_EMPTY_COLS)
        expected = np.array(
            [
                [0.4, 0.4, 0.0, 0.2],
                [np.nan, np.nan, np.nan, np.nan],
                [0.0, 0.33333333, 0.0, 0.66666667],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, 0.66666667, 0.0, 0.33333333],
                [0.0, 1.0, 0.0, 0.0],
            ]
        )
        actual = cube.proportions(axis=1, prune=False)
        expected = np.array(
            [
                [0.4, 0.4, 0.2],
                [0.0, 0.33333333, 0.66666667],
                [0.0, 1.0, 0.0],
                [0.0, 0.66666667, 0.33333333],
                [0.0, 1.0, 0.0],
            ]
        )
        table = cube.proportions(axis=1, prune=True)
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_almost_equal(actual, expected)

        pruned_expected = [
            np.array([False, True, False, False, False, False]),
            np.array([False, False, True, False]),
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(pruned[i], pruned_expected[i])

    def test_cat_x_cat_props_by_cell_prune_cols(self):
        cube = CrunchCube(CR.CAT_X_CAT_WITH_EMPTY_COLS)
        expected = np.array(
            [
                [0.14285714, 0.14285714, 0.0, 0.07142857],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.07142857, 0.0, 0.14285714],
                [0.0, 0.14285714, 0.0, 0.0],
                [0.0, 0.14285714, 0.0, 0.07142857],
                [0.0, 0.07142857, 0.0, 0.0],
            ]
        )
        actual = cube.proportions(axis=None, prune=False)
        expected = np.array(
            [
                [0.14285714, 0.14285714, 0.07142857],
                [0.0, 0.07142857, 0.14285714],
                [0.0, 0.14285714, 0.0],
                [0.0, 0.14285714, 0.07142857],
                [0.0, 0.07142857, 0.0],
            ]
        )
        table = cube.proportions(axis=None, prune=True)
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_almost_equal(actual, expected)

        pruned_expected = [
            np.array([False, True, False, False, False, False]),
            np.array([False, False, True, False]),
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(pruned[i], pruned_expected[i])

    def test_cat_x_cat_index_by_col_prune_cols(self):
        cube = CrunchCube(CR.CAT_X_CAT_WITH_EMPTY_COLS)
        expected = np.array(
            [
                [2.8, 0.7, np.nan, 0.7],
                [np.nan, np.nan, np.nan, np.nan],
                [0.0, 0.58333333, np.nan, 2.33333333],
                [0.0, 1.75, np.nan, 0.0],
                [0.0, 1.16666667, np.nan, 1.16666667],
                [0.0, 1.75, np.nan, 0.0],
            ]
        )
        actual = Index.data(cube, weighted=True, prune=False)
        # Assert index without pruning
        np.testing.assert_almost_equal(actual, expected)

        expected = np.array(
            [
                [2.8, 0.7, 0.7],
                [0.0, 0.58333333, 2.33333333],
                [0.0, 1.75, 0.0],
                [0.0, 1.16666667, 1.16666667],
                [0.0, 1.75, 0.0],
            ]
        )
        table = Index.data(cube, weighted=True, prune=True)
        # Assert index witih pruning
        actual = table[:, ~table.mask.all(axis=0)][~table.mask.all(axis=1), :]
        np.testing.assert_almost_equal(actual, expected)

        pruned_expected = [
            np.array([False, True, False, False, False, False]),
            np.array([False, False, True, False]),
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(pruned[i], pruned_expected[i])

    def test_prune_univariate_cat(self):
        cube = CrunchCube(CR.BINNED)
        expected = np.array([118504.40402204, 155261.2723631, 182923.95470245])
        actual = cube.as_array(prune=True)
        np.testing.assert_almost_equal(actual[~actual.mask], expected)

        pruned_expected = [
            np.array(
                [
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                ]
            )
        ]
        pruned = cube.prune_indices()
        self.assertEqual(len(pruned), len(pruned_expected))
        for i, actual in enumerate(pruned):
            np.testing.assert_array_equal(pruned[i], pruned_expected[i])

    def test_single_col_margin_not_iterable(self):
        cube = CrunchCube(CR.SINGLE_COL_MARGIN_NOT_ITERABLE)
        expected = (1,)
        actual = cube.margin(axis=0).shape
        self.assertEqual(actual, expected)

    def test_3d_percentages_by_col(self):
        # ---CAT x CAT x CAT---
        cube = CrunchCube(CR.GENDER_PARTY_RACE)
        expected = np.array(
            [
                [
                    [0.17647059, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.17647059, 0.05882353, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.23529412, 0.0, 0.0, 0.0, 0.0, 0.05882353, 0.0, 0.0],
                    [
                        0.11764706,
                        0.05882353,
                        0.0,
                        0.05882353,
                        0.0,
                        0.05882353,
                        0.0,
                        0.0,
                    ],
                ],
                [
                    [0.04761905, 0.0, 0.0, 0.04761905, 0.0, 0.0, 0.0, 0.0],
                    [
                        0.14285714,
                        0.04761905,
                        0.0952381,
                        0.04761905,
                        0.0,
                        0.04761905,
                        0.0,
                        0.0,
                    ],
                    [0.23809524, 0.0, 0.04761905, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.19047619, 0.0, 0.04761905, 0.0, 0.0, 0.0, 0.0, 0.0],
                ],
            ]
        )
        # Set axis to tuple (1, 2), since we want to do a total for each slice
        # of the 3D cube. This is consistent with how the np.sum works
        # (axis parameter), which is used internally in
        # 'proportions' calculation.
        axis = (1, 2)
        actual = cube.proportions(axis=axis)
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_dim_ind_for_cat_cube(self):
        cube = CrunchCube(CR.CAT_X_CAT)
        expected = None
        actual = cube.mr_dim_ind
        self.assertEqual(actual, expected)

    def test_total_unweighted_margin_when_has_means(self):
        """Tests that total margin is Unweighted N, when cube has means."""
        cube = CrunchCube(CR.CAT_MEAN_WGTD)
        expected = 17615
        actual = cube.margin(weighted=False)
        assert actual == expected

    def test_row_unweighted_margin_when_has_means(self):
        """Tests that total margin is Unweighted N, when cube has means."""
        cube = CrunchCube(CR.CAT_MEAN_WGTD)
        expected = np.array(
            [
                806,
                14,
                14,
                28,
                780,
                42,
                1114,
                28,
                24,
                746,
                2,
                12,
                6,
                2178,
                2026,
                571,
                136,
                16,
                14,
                1334,
                1950,
                26,
                128,
                4,
                28,
                3520,
                1082,
                36,
                56,
                556,
                38,
                146,
                114,
                28,
                12,
            ]
        )
        actual = np.ma.compressed(cube.margin(axis=1, weighted=False, prune=True))
        np.testing.assert_array_equal(actual, expected)
        # not testing cube.prune_indices() because the margin has 6367 cells

    def test_ca_with_single_cat_pruning(self):
        cube = CrunchCube(CR.CA_SINGLE_CAT)
        # The last 0 of the expectation is not visible in whaam because of
        # pruning, which is not the responsibility of cr.cube.
        expected = np.array([79, 80, 70])
        actual = np.ma.compressed(cube.as_array(weighted=False, prune=True))
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_single_cat_counts(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)

        expected = np.array([[13, 12], [16, 12], [11, 12]])
        actual = cube.as_array()
        np.testing.assert_array_equal(actual, expected)

        # Expectation for pruned indices (none are pruned)
        expecteds = [
            (np.array([False, False]), np.array([False])),
            (np.array([False, False]), np.array([False])),
            (np.array([False, False]), np.array([False])),
        ]
        actuals = cube.prune_indices()
        for expected, actual in zip(expecteds, actuals):
            np.testing.assert_array_equal(actual[0], expected[0])
            np.testing.assert_array_equal(actual[1], expected[1])

    def test_ca_x_single_cat_props_by_col(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)
        expected = np.array(
            [[0.52, 0.48], [0.57142857, 0.42857143], [0.47826087, 0.52173913]]
        )
        # Col direction is 1 (and not 0) because of 3D cube.
        actual = cube.proportions(axis=1)
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_x_single_cat_props_by_row(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)
        expected = np.array([[1.0, 1.0], [1.0, 1.0], [1.0, 1.0]])
        # Col direction is 2 (and not 1) because of 3D cube.
        actual = cube.proportions(axis=2)
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_x_single_cat_props_by_cell(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)
        expected = np.array(
            [[0.52, 0.48], [0.57142857, 0.42857143], [0.47826087, 0.52173913]]
        )
        # Col direction is (1, 2) because 3D cube (total per slice).
        actual = cube.proportions(axis=(1, 2))
        np.testing.assert_almost_equal(actual, expected)

    def test_ca_x_single_cat_col_margins(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)
        expected = np.array([25, 28, 23])
        # Axis equals to 1, because col direction in 3D cube is 1 (and not 0).
        # It operates on the 0th dimension of each slice (which is 1st
        # dimension of the cube).
        actual = cube.margin(axis=1)
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_single_cat_row_margins(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)
        expected = np.array([[13, 12], [16, 12], [11, 12]])
        # Axis equals to 2, because col direction in 3D cube is 2 (and not 1).
        # It operates on the 1st dimension of each slice (which is 2nd
        # dimension of the cube).
        actual = cube.margin(axis=2)
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_single_cat_cell_margins(self):
        cube = CrunchCube(CR.CA_X_SINGLE_CAT)
        expected = np.array([25, 28, 23])
        # Axis equals to (1, 2), because the total is calculated for each slice.
        actual = cube.margin(axis=(1, 2))
        np.testing.assert_array_equal(actual, expected)

    def test_ca_subvar_x_cat_hs_counts_prune(self):
        cube = CrunchCube(CR.CA_SUBVAR_X_CAT_HS)
        expected = np.array([[3, 3, 0, 0, 6], [1, 3, 2, 0, 4], [0, 2, 1, 3, 2]])
        actual = cube.as_array(include_transforms_for_dims=[0, 1], prune=True)
        np.testing.assert_array_equal(actual, expected)

    def test_means_univariate_cat(self):
        cube = CrunchCube(CR.ECON_BLAME_WITH_HS)
        expected = [[np.array(2.1735205616850553)]]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_bivariate_cat(self):
        cube = CrunchCube(CR.ECON_BLAME_X_IDEOLOGY_ROW_HS)
        expected = [
            [
                np.array(
                    [2.19444444, 2.19230769, 2.26666667, 1.88990826, 1.76363636, 3.85]
                )
            ]
        ]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_cat_x_mr(self):
        cube = CrunchCube(CR.FRUIT_X_PETS)
        expected = [[np.array([1.7, 1.6470588, 1.6842105]), None]]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_mr_x_cat(self):
        cube = CrunchCube(CR.PETS_X_FRUIT)
        expected = [[None, np.array([1.7, 1.6470588, 1.6842105])]]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_cat_array_cat_dim_first(self):
        cube = CrunchCube(CR.PETS_ARRAY_CAT_FIRST)
        expected = [[None, np.array([1.44333002, 1.48049069, 1.57881177])]]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_cat_array_subvar_dim_first(self):
        cube = CrunchCube(CR.PETS_ARRAY_SUBVAR_FIRST)
        expected = [[np.array([1.44333002, 1.48049069, 1.57881177]), None]]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_cat_x_cat_arr_fruit_first(self):
        cube = CrunchCube(CR.FRUIT_X_PETS_ARRAY)
        expected = [
            [None, np.array([1.48, 1.42857143, 1.52173913])],
            [None, np.array([1.40740741, 1.53846154, 1.55319149])],
        ]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_cat_x_cat_arr_subvars_first(self):
        cube = CrunchCube(CR.FRUIT_X_PETS_ARRAY_SUBVARS_FIRST)
        expected = [
            [np.array([1.71111111, 1.6, 1.65625]), None],
            [np.array([1.64705882, 1.7, 1.68421053]), None],
        ]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_cat_x_cat_arr_pets_first(self):
        cube = CrunchCube(CR.FRUIT_X_PETS_ARRAY_PETS_FIRST)
        expected = [
            [np.array([1.48, 1.40740741]), np.array([1.71111111, 1.64705882])],
            [np.array([1.42857143, 1.53846154]), np.array([1.6, 1.7])],
            [np.array([1.52173913, 1.55319149]), np.array([1.65625, 1.68421053])],
        ]
        actual = cube.scale_means()
        assert_scale_means_equal(actual, expected)

    def test_means_with_null_values(self):
        cube = CrunchCube(CR.SCALE_WITH_NULL_VALUES)
        scale_means = cube.scale_means()
        assert_scale_means_equal(
            scale_means, [[np.array([1.2060688, 1.0669344, 1.023199]), None]]
        )

    def test_values_services(self):
        cube = CrunchCube(CR.MR_X_CA_CAT_X_CA_SUBVAR)
        # Axis is 1, which is 'col' direction, since 3D cube.
        actual = cube.proportions(axis=1)[0]
        expected = np.array(
            [
                [
                    0.14285714,
                    0.10204082,
                    0.20512821,
                    0.16363636,
                    0.16438356,
                    0.1372549,
                    0.18181818,
                    0.2991453,
                    0.32,
                    0.44776119,
                ],  # noqa
                [
                    0.07142857,
                    0.23469388,
                    0.17948718,
                    0.14545455,
                    0.20547945,
                    0.09803922,
                    0.27272727,
                    0.11111111,
                    0.352,
                    0.23880597,
                ],  # noqa
                [
                    0.12857143,
                    0.19387755,
                    0.1025641,
                    0.16363636,
                    0.1369863,
                    0.15686275,
                    0.25,
                    0.17094017,
                    0.136,
                    0.14925373,
                ],  # noqa
                [
                    0.15714286,
                    0.15306122,
                    0.14102564,
                    0.05454545,
                    0.17808219,
                    0.09803922,
                    0.18181818,
                    0.20512821,
                    0.064,
                    0.05223881,
                ],  # noqa
                [
                    0.12857143,
                    0.12244898,
                    0.1025641,
                    0.05454545,
                    0.15068493,
                    0.07843137,
                    0.06060606,
                    0.1025641,
                    0.064,
                    0.05970149,
                ],  # noqa
                [
                    0.05714286,
                    0.09183673,
                    0.20512821,
                    0.09090909,
                    0.09589041,
                    0.11764706,
                    0.03030303,
                    0.02564103,
                    0.032,
                    0.01492537,
                ],  # noqa
                [
                    0.08571429,
                    0.04081633,
                    0.05128205,
                    0.07272727,
                    0.01369863,
                    0.11764706,
                    0.01515152,
                    0.05128205,
                    0.024,
                    0.02238806,
                ],  # noqa
                [
                    0.17142857,
                    0.04081633,
                    0.01282051,
                    0.03636364,
                    0.02739726,
                    0.01960784,
                    0.00757576,
                    0.00854701,
                    0.008,
                    0.00746269,
                ],  # noqa
                [
                    0.01428571,
                    0.02040816,
                    0.0,
                    0.14545455,
                    0.01369863,
                    0.11764706,
                    0.0,
                    0.0,
                    0.0,
                    0.00746269,
                ],  # noqa
                [
                    0.04285714,
                    0.0,
                    0.0,
                    0.07272727,
                    0.01369863,
                    0.05882353,
                    0.0,
                    0.02564103,
                    0.0,
                    0.0,
                ],  # noqa
                [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )
        np.testing.assert_almost_equal(actual, expected)
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_props_with_hs_by_cell(self):
        cube = CrunchCube(CR.LETTERS_X_PETS_HS)
        expected = np.array(
            [
                [
                    0.10769231,
                    0.16923077,
                    0.27692308,
                    0.26153846,
                    0.15384615,
                    0.15384615,
                ],  # noqa
                [
                    0.11111111,
                    0.20634921,
                    0.31746032,
                    0.19047619,
                    0.15873016,
                    0.15873016,
                ],  # noqa
                [
                    0.09090909,
                    0.22727273,
                    0.31818182,
                    0.24242424,
                    0.12121212,
                    0.12121212,
                ],  # noqa
                [
                    0.10447761,
                    0.14925373,
                    0.25373134,
                    0.13432836,
                    0.17910448,
                    0.17910448,
                ],  # noqa
                [
                    0.07462687,
                    0.11940299,
                    0.19402985,
                    0.23880597,
                    0.1641791,
                    0.1641791,
                ],  # noqa
            ]
        )
        actual = cube.proportions(include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_props_with_hs_by_row(self):
        cube = CrunchCube(CR.LETTERS_X_PETS_HS)
        expected = np.array(
            [
                [0.15555556, 0.24444444, 0.4, 0.37777778, 0.22222222, 0.22222222],
                [
                    0.16666667,
                    0.30952381,
                    0.47619048,
                    0.28571429,
                    0.23809524,
                    0.23809524,
                ],  # noqa
                [
                    0.13333333,
                    0.33333333,
                    0.46666667,
                    0.35555556,
                    0.17777778,
                    0.17777778,
                ],  # noqa
                [
                    0.18421053,
                    0.26315789,
                    0.44736842,
                    0.23684211,
                    0.31578947,
                    0.31578947,
                ],  # noqa
                [0.125, 0.2, 0.325, 0.4, 0.275, 0.275],
            ]
        )
        actual = cube.proportions(axis=1, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_mr_props_with_hs_by_col(self):
        cube = CrunchCube(CR.LETTERS_X_PETS_HS)
        expected = np.array(
            [
                [
                    0.53846154,
                    0.6875,
                    0.62068966,
                    0.94444444,
                    0.55555556,
                    0.55555556,
                ],  # noqa
                [
                    0.58333333,
                    0.68421053,
                    0.64516129,
                    0.66666667,
                    0.71428571,
                    0.71428571,
                ],  # noqa
                [0.5, 0.78947368, 0.67741935, 0.76190476, 0.57142857, 0.57142857],
                [0.53846154, 0.58823529, 0.56666667, 0.5, 0.63157895, 0.63157895],
                [
                    0.45454545,
                    0.47058824,
                    0.46428571,
                    0.76190476,
                    0.61111111,
                    0.61111111,
                ],  # noqa
            ]
        )
        actual = cube.proportions(axis=0, include_transforms_for_dims=[0, 1])
        np.testing.assert_almost_equal(actual, expected)

    def test_3d_pruning_indices(self):
        """Test pruning indices for a simple XYZ cube."""
        cube = CrunchCube(CR.XYZ_SIMPLE_ALLTYPES)

        # Zeroth slice of the XYZ array:
        #
        # +----+-----+----+
        # |  0 |  0  | 0  | True
        # +----+-----+----+
        # |  0 |  0  | 0  | True
        # +----+-----+----+
        # |  0 |  1  | 0  | False
        # +----+-----+----+
        # |  0 |  0  | 0  | True
        # +----+-----+----+
        # |  0 |  0  | 0  | True
        # +----+-----+----+
        #  True False True

        expected = (
            np.array([True, True, False, True, True]),
            np.array([True, False, True]),
        )
        actual = cube.prune_indices()[0]
        np.testing.assert_array_equal(actual[0], expected[0])
        np.testing.assert_array_equal(actual[1], expected[1])

    def test_mr_x_ca_rows_margin(self):
        cube = CrunchCube(CR.MR_X_CA_HS)
        actual = cube.margin(axis=2)
        expected = np.array([[3, 3, 3], [4, 4, 4], [0, 0, 0]])
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_mr_margin(self):
        cube = CrunchCube(CR.CA_X_MR_WEIGHTED_HS)
        expected = np.array([504, 215, 224, 76, 8, 439])
        actual = cube.margin(
            axis=1, weighted=False, include_transforms_for_dims=[0, 1, 2]
        )[0]
        np.testing.assert_array_equal(actual, expected)

    def test_ca_x_mr_margin_prune(self):
        # ---CA x MR---
        cube = CrunchCube(CR.CA_X_MR_WEIGHTED_HS)

        margin = cube.margin(
            axis=1, weighted=False, include_transforms_for_dims=[0, 1, 2], prune=True
        )[0]

        np.testing.assert_array_equal(margin, np.array([504, 215, 224, 76, 8, 439]))

    def test_mr_x_cat_x_mr_pruning(self):
        cube = CrunchCube(CR.MR_X_CAT_X_MR_PRUNE)
        expected = np.array(
            [
                [False, False, False, True],
                [False, False, False, True],
                [True, True, True, True],
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
                [True, True, True, True],
                [True, True, True, True],
            ]
        )
        actual = cube.proportions(prune=True)[0].mask
        np.testing.assert_array_equal(actual, expected)

    def test_mr_x_mr_prune_indices(self):
        cube = CrunchCube(CR.HUFFPOST_ACTIONS_X_HOUSEHOLD)
        expected = [
            np.array([False, False, False, False, False, False, False, False]),
            np.array([False, False, False]),
        ]
        actual = cube.prune_indices()
        np.testing.assert_array_equal(actual[0], expected[0])
        np.testing.assert_array_equal(actual[1], expected[1])

    def test_mr_x_cat_hs_prune_indices(self):
        cube = CrunchCube(CR.MR_X_CAT_HS)
        expected = [
            np.array([False, False, False, False, False]),
            np.array([False, False, False, True, False, False, True, False]),
        ]
        actual = cube.prune_indices(transforms=[0, 1])
        np.testing.assert_array_equal(actual[0], expected[0])
        np.testing.assert_array_equal(actual[1], expected[1])

    def test_gender_x_weight_pruning(self):
        cube = CrunchCube(CR.GENDER_X_WEIGHT)
        expected = 208
        actual = cube.margin(prune=True)
        np.testing.assert_array_equal(actual, expected)

    def test_proportions_cat_x_mr_x_cat(self):
        cube = CrunchCube(CR.CAT_X_MR_X_CAT["slides"][0]["cube"])
        props = cube.proportions(axis=1, include_transforms_for_dims=[1, 2], prune=True)

        # Test first slice
        expected = np.array(
            [
                [0.3647320622384293, 0.3539601689446188],
                [0.2217369551352075, 0.21179394663351137],
                [0.2892943293978049, 0.2755799995757032],
                [0.30472468951159837, 0.31183395927247587],
                [0.1564932915102434, 0.12122561350397994],
                [0.14341569357975512, 0.16573829062583212],
                [0.3088773171830892, 0.30399159711003093],
                [0.3531835200406305, 0.4001713397700346],
                [0.2572968713520773, 0.24375187975969445],
                [0.25130939319998297, 0.1977549624477041],
                [0.274565755319545, 0.2997170882672239],
                [0.32608867343144654, 0.344478827337916],
                [0.35921238787890847, 0.3513630660099009],
                [0.25634234169007164, 0.16894820580901035],
                [0.22099299650777843, 0.22466833670878553],
                [0.2276649518247445, 0.24565762260105783],
                [0.2643381997593724, 0.1982327504994244],
                [0.41491884119922506, 0.4258666971460735],
                [0.4623019656879477, 0.520868963921971],
                [0.35277296769118416, 0.2813894529707485],
                [0.4003762375617921, 0.42681767440587187],
                [0.25925087940264385, 0.2731916153395818],
                [0.4753330440584336, 0.43648850969829933],
                [0.28148528548474344, 0.24591762645117995],
                [0.49727220036895625, 0.5108530445282087],
                [0.32590772578924143, 0.359683660082846],
                [0.3595152413326164, 0.4049397928654146],
                [0.42108775668830034, 0.3454454870295823],
                [0.4205886117048783, 0.39122538735554757],
                [0.228770284468049, 0.20059146301631123],
                [0.24603034757739972, 0.2735752782805667],
                [0.3065494749862463, 0.32314540506421324],
                [0.27194030884117687, 0.30970380251451973],
                [0.2091262494207975, 0.22920517801957993],
                [0.31769765653105564, 0.28958158962721786],
                [0.3560762345705489, 0.3102687419303191],
                [0.4804715042882989, 0.5011391743289532],
                [0.3811599892254701, 0.4226226669501276],
                [0.41029213392178276, 0.1884401015918774],
            ]
        )
        actual = compress_pruned(props[0])
        np.testing.assert_almost_equal(actual, expected)

        # Test second slice
        expected = np.array(
            [
                [0.4031214008509537, 0.4056176483118717],
                [0.24070198556407071, 0.25911741489783324],
                [0.3464300357360538, 0.33912070707907394],
                [0.3574620243397532, 0.37758989006965377],
                [0.14907456320910992, 0.15890701660953999],
                [0.20651100193920027, 0.1789634776591901],
                [0.30520247457440536, 0.3270678407142803],
                [0.3810285430516052, 0.3997309090998591],
                [0.34271319381187965, 0.17385655771473044],
                [0.3591305867483556, 0.2685531153107514],
                [0.2996020578719067, 0.29215934221779744],
                [0.3902291806673195, 0.39354067543725346],
                [0.40337866312317217, 0.4250345472210801],
                [0.32114152359818676, 0.30423847092340256],
                [0.2653311867224287, 0.27833063737964403],
                [0.313528046383254, 0.3000437600045656],
                [0.3920027268884396, 0.36933090371694116],
                [0.515781215492543, 0.4851941444303692],
                [0.5427173809468948, 0.5754703450096686],
                [0.3514753251895221, 0.3381463298582681],
                [0.40789566838261765, 0.43525540615386243],
                [0.3595684862225866, 0.38051785122829174],
                [0.556336290160639, 0.47918668411102733],
                [0.3006536550040242, 0.31770376362899333],
                [0.5132046910093269, 0.5548090283383379],
                [0.4409089206826546, 0.36616426510098665],
                [0.40390891699096854, 0.3888593109712533],
                [0.3914326716352874, 0.3346157095319079],
                [0.4423415709934932, 0.47752483308415145],
                [0.33543493750667275, 0.30087121770598385],
                [0.30409560492337334, 0.33096009035672747],
                [0.4028029720384424, 0.4497865293548307],
                [0.37991282964691514, 0.3817002730808065],
                [0.26937198075202085, 0.2530238979016483],
                [0.3367641290249356, 0.3210357156509789],
                [0.4376473666508847, 0.4044796591984694],
                [0.5986306705327854, 0.5886692367162286],
                [0.3493779725965881, 0.3533483607971598],
                [0.08962504168280223, 0.4352830423033842],
            ]
        )
        actual = compress_pruned(props[1])
        np.testing.assert_almost_equal(actual, expected)

    def test_univ_mr_with_hs_does_not_crash(self):
        """Assert that MR with H&S doesn't crash."""
        cube = CrunchCube(CR.UNIV_MR_WITH_HS["slides"][0]["cube"])
        cube.as_array(include_transforms_for_dims=[0])
        # If it doesn't crash, the test passes, we don't actually care about
        # the result. We only care that the H&S transform doesn't crash the MR
        # variable (that it doesn't try to actually include the transform
        # in the result). H&S shouldn't be in the MR variable, but there
        # are cases when there are.
        assert True

    def test_pop_counts_ca_as_0th(self):
        cube_slice = CrunchCube(CR.CA_AS_0TH).get_slices(ca_as_0th=True)[0]
        expected = np.array(
            [54523323.46453754, 24570078.10865863, 15710358.25446403, 5072107.27712256]
        )
        actual = cube_slice.population_counts(100000000)
        np.testing.assert_almost_equal(actual, expected)

    def test_pop_counts_for_multiple_slices(self):
        cube = CrunchCube(CR.PETS_ARRAY_X_PETS)
        expected = np.array(
            [
                [
                    [24992768.29621058, 0.0, 26901938.09661558],
                    [17298235.46427536, 44258027.19120625, 21174428.69540066],
                ],
                [
                    [0.0, 19459910.91314028, 24053452.11581288],
                    [48106904.23162583, 16648106.90423161, 22216035.63474388],
                ],
                [
                    [21474773.60931439, 18272962.4838292, 0.0],
                    [25808538.16300128, 23673997.41267791, 53751617.07632601],
                ],
            ]
        )
        actual = cube.population_counts(100000000)
        np.testing.assert_almost_equal(actual, expected)



# encoding: utf-8

"""T-score based P-values of pairwise comparison or columns of a contingency table."""

from __future__ import division

import numpy as np
from scipy.stats import t

from cr.cube.util import lazyproperty

try:
    xrange
except NameError:  # pragma: no cover
    # pylint: disable=invalid-name
    xrange = range


class NewPairwiseSignificance:
    """Implementation of p-vals and t-tests for each column proportions comparison."""

    def __init__(self, slice_, alpha=0.05, only_larger=True):
        self._slice = slice_
        self._alpha = alpha
        self._only_larger = only_larger

    @lazyproperty
    def values(self):
        """list of _ColumnPairwiseSignificance tests.

        Result has as many elements as there are coliumns in the slice. Each
        significance test contains `p_vals` and `t_stats` significance tests.
        """
        return [
            _ColumnPairwiseSignificance(
                self._slice, col_idx, self._alpha, self._only_larger
            )
            for col_idx in xrange(self._slice.shape[1])
        ]

    @lazyproperty
    def pairwise_indices(self):
        """ndarray containing tuples of pairwise indices."""
        return np.array([sig.pairwise_indices for sig in self.values]).T

    @lazyproperty
    def summary_pairwise_indices(self):
        """ndarray containing tuples of pairwise indices for the column summary."""
        summary_pairwise_indices = np.empty(
            self.values[0].t_stats.shape[1], dtype=object
        )
        summary_pairwise_indices[:] = [
            sig.summary_pairwise_indices for sig in self.values
        ]
        return summary_pairwise_indices


class _ColumnPairwiseSignificance:
    """Value object providing matrix of T-score based pairwise-comparison P-values"""

    def __init__(self, slice_, col_idx, alpha=0.05, only_larger=True):
        self._slice = slice_
        self._col_idx = col_idx
        self._alpha = alpha
        self._only_larger = only_larger

    @lazyproperty
    def t_stats(self):
        props = self._slice.column_proportions
        diff = props - props[:, [self._col_idx]]
        var_props = props * (1.0 - props) / self._slice.column_base
        se_diff = np.sqrt(var_props + var_props[:, [self._col_idx]])
        return diff / se_diff

    @lazyproperty
    def p_vals(self):
        return 2 * (1 - t.cdf(abs(self.t_stats), df=self._df))

    @lazyproperty
    def pairwise_indices(self):
        significance = self.p_vals < self._alpha
        if self._only_larger:
            significance = np.logical_and(self.t_stats < 0, significance)
        return [tuple(np.where(sig_row)[0]) for sig_row in significance]

    @lazyproperty
    def summary_pairwise_indices(self):
        significance = self.summary_p_vals < self._alpha
        if self._only_larger:
            significance = np.logical_and(self.summary_t_stats < 0, significance)
        return tuple(np.where(significance)[0])

    @lazyproperty
    def summary_t_stats(self):
        col_margin_props = self._slice.column_base / self._slice.table_margin
        diff = col_margin_props - col_margin_props[self._col_idx]
        var_props = (
            col_margin_props * (1.0 - col_margin_props) / self._slice.table_margin
        )
        se_diff = np.sqrt(var_props + var_props[self._col_idx])
        return diff / se_diff

    @lazyproperty
    def summary_p_vals(self):
        return 2 * (1 - t.cdf(abs(self.summary_t_stats), df=self._df))

    @lazyproperty
    def _df(self):
        selected_unweighted_n = (
            self._slice.column_base[self._col_idx]
            if self._slice.column_base.ndim < 2
            else self._slice.column_base[:, self._col_idx][:, None]
        )
        return self._slice.column_base + selected_unweighted_n - 2



# encoding: utf-8

"""Provides the CrunchCube class.

CrunchCube is the main API class for manipulating Crunch.io JSON cube
responses.
"""

from __future__ import division

import json
import warnings

import numpy as np

from cr.cube.cube_slice import CubeSlice
from cr.cube.dimension import AllDimensions
from cr.cube.enum import DIMENSION_TYPE as DT
from cr.cube.measures.index import Index
from cr.cube.measures.scale_means import ScaleMeans
from cr.cube.util import lazyproperty

np.seterr(divide="ignore", invalid="ignore")


class CrunchCube(object):
    """The main API object for manipulating Crunch.io cube-responses.

    This object provides the main API methods for working with cubes. The
    main API functions are:

    * `as_array()`
    * `margin()`
    * `proportions()`
    * `percentages()`

    These methods return statistical information of interest from a JSON
    cube response.

    Crunch Cubes contain richer metadata than standard Python objects and
    also conceal certain complexity in the data structures from the user.
    In particular, multiple-response variables are generally represented as
    single dimensions in result tables, but in the actual data, they may
    comprise of two dimensions. These methods (API) understand the subtleties
    in the Crunch data types, and correctly compute margins and percentages
    from them.
    """

    def __init__(self, response):
        """Initializes the CrunchCube class with the cube JSON response.

        Class can be initialized with both JSON string, and dict types.
        There's no real parsing of the data at the point of initialization,
        just storing. The functionality is implemented in the
        appropriate API functions.

        @response: Represents the cube response object, as generated by the
            zz9d cube factory. Cube can come in as a JSON or as a dictionary,
            So we need to check its type, and convert it to a dictionary if
            it's JSON, if possible.
        """
        self._cube_response_arg = response

    def __repr__(self):
        """Provide text representation suitable for working at console.

        Falls back to a default repr on exception, such as might occur in
        unit tests where object need not otherwise be provided with all
        instance variable values.
        """
        try:
            dimensionality = " x ".join(dt.name for dt in self.dim_types)
            return "%s(name='%s', dim_types='%s')" % (
                type(self).__name__,
                self.name,
                dimensionality,
            )
        except Exception:
            return super(CrunchCube, self).__repr__()

    def as_array(
        self,
        include_missing=False,
        weighted=True,
        include_transforms_for_dims=None,
        prune=False,
    ):
        """Return `ndarray` representing cube values.

        Returns the tabular representation of the crunch cube. The returned
        array has the same number of dimensions as the cube. E.g. for
        a cross-tab representation of a categorical and numerical variable,
        the resulting cube will have two dimensions.

        *include_missing* (bool): Include rows/cols for missing values.

        Example 1 (Categorical x Categorical)::

            >>> cube = CrunchCube(response)
            >>> cube.as_array()
            np.array([
                [5, 2],
                [5, 3],
            ])

        Example 2 (Categorical x Categorical, include missing values)::

            >>> cube = CrunchCube(response)
            >>> cube.as_array(include_missing=True)
            np.array([
                [5, 3, 2, 0],
                [5, 2, 3, 0],
                [0, 0, 0, 0],
            ])
        """
        array = self._as_array(
            include_missing=include_missing,
            weighted=weighted,
            include_transforms_for_dims=include_transforms_for_dims,
        )

        # ---prune array if pruning was requested---
        if prune:
            array = self._prune_body(array, transforms=include_transforms_for_dims)

        return self._drop_mr_cat_dims(array)

    @lazyproperty
    def can_compare_pairwise(self):
        return all(slice_.can_compare_pairwise for slice_ in self.slices)

    def count(self, weighted=True):
        """Return numberic count of rows considered for cube response."""
        return self._measures.weighted_n if weighted else self._measures.unweighted_n

    @lazyproperty
    def description(self):
        """Return the description of the cube."""
        if not self.dimensions:
            return None
        return self.dimensions[0].description

    @lazyproperty
    def dim_types(self):
        """Tuple of DIMENSION_TYPE member for each dimension of cube."""
        return tuple(d.dimension_type for d in self.dimensions)

    @lazyproperty
    def dimensions(self):
        """_ApparentDimension object providing access to visible dimensions.

        A cube involving a multiple-response (MR) variable has two dimensions
        for that variable (subvariables and categories dimensions), but is
        "collapsed" into a single effective dimension for cube-user purposes
        (its categories dimension is supressed). This collection will contain
        a single dimension for each MR variable and therefore may have fewer
        dimensions than appear in the cube response.
        """
        return self._all_dimensions.apparent_dimensions

    @lazyproperty
    def filter_annotation(self):
        """Get cube's filter annotation."""
        return self._cube_dict.get("filter_names", [])

    def get_slices(self, ca_as_0th=False):
        """Return list of :class:`.CubeSlice` objects.

        The number of slice objects in the returned list depends on the
        dimensionality of this cube. A 1D or 2D cube will return a list
        containing one slice object. A 3D cube will return a list of slices
        the same length as the first dimension.
        """
        if self.ndim < 3 and not ca_as_0th:
            return [CubeSlice(self, 0)]

        return [CubeSlice(self, i, ca_as_0th) for i, _ in enumerate(self.labels()[0])]

    @lazyproperty
    def has_means(self):
        """True if cube includes a means measure."""
        return self._measures.means is not None

    @lazyproperty
    def has_mr(self):
        """True if this cube has a multiple-response dimension."""
        return self.mr_dim_ind is not None

    def index(self, weighted=True, prune=False):
        """Return cube index measurement.

        This function is deprecated. Use index_table from CubeSlice.
        """
        warnings.warn(
            "CrunchCube.index() is deprecated. Use CubeSlice.index_table().",
            DeprecationWarning,
        )
        return Index.data(self, weighted, prune)

    def inserted_hs_indices(self, prune=False):
        """Get indices of the inserted H&S (for formatting purposes)."""
        if self.ndim == 2 and prune:
            # If pruning is applied, we need to subtract from the H&S indes
            # the number of pruned rows (cols) that come before that index.
            pruning_bases = [self._pruning_base(axis=i, hs_dims=[0, 1]) for i in [1, 0]]
            pruning_bases = [
                base if base.ndim == 1 else np.sum(base, axis=(1 - i))
                for i, base in enumerate(pruning_bases)
            ]
            # Obtain prune indices as subscripts
            prune_indices_list = [
                np.arange(len(base))[np.logical_or(base == 0, np.isnan(base))]
                for base in pruning_bases
            ]
            inserted_indices_list = [dim.inserted_hs_indices for dim in self.dimensions]
            return self._adjust_inserted_indices(
                inserted_indices_list, prune_indices_list
            )

        return [dim.inserted_hs_indices for dim in self.dimensions]

    @lazyproperty
    def is_double_mr(self):
        """Check if cube has 2 MR dimensions."""
        return True if isinstance(self.mr_dim_ind, tuple) else False

    @lazyproperty
    def is_univariate_ca(self):
        """True if cube only contains a CA dimension-pair, in either order."""
        return self.ndim == 2 and set(self.dim_types) == {DT.CA_SUBVAR, DT.CA_CAT}

    @lazyproperty
    def is_weighted(self):
        """True if cube response contains weighted data."""
        return self._measures.is_weighted

    def labels(self, include_missing=False, include_transforms_for_dims=False):
        """Gets labels for each cube's dimension.

        Args
            include_missing (bool): Include labels for missing values

        Returns
            labels (list of lists): Labels for each dimension
        """
        return [
            dim.labels(include_missing, include_transforms_for_dims)
            for dim in self.dimensions
        ]

    def margin(
        self,
        axis=None,
        weighted=True,
        include_missing=False,
        include_transforms_for_dims=None,
        prune=False,
        include_mr_cat=False,
    ):
        """Get margin for the selected axis.

        the selected axis. For MR variables, this is the sum of the selected
        and non-selected slices.

        Args
            axis (int): Axis across the margin is calculated. If no axis is
                        provided the margin is calculated across all axis.
                        For Categoricals, Num, Datetime, and Text, this
                        translates to sumation of all elements.
        Returns
            Calculated margin for the selected axis

        Example 1:
            >>> cube = CrunchCube(fixt_cat_x_cat)
            np.array([
               [5, 2],
               [5, 3],
            ])

            >>> cube.margin(axis=0)
            np.array([10, 5])

        Example 2:
            >>> cube = CrunchCube(fixt_cat_x_num_x_datetime)
            np.array([
                [[1, 1],
                 [0, 0],
                 [0, 0],
                 [0, 0]],
                [[2, 1],
                 [1, 1],
                 [0, 0],
                 [0, 0]],
                [[0, 0],
                 [2, 3],
                 [0, 0],
                 [0, 0]],
                [[0, 0],
                 [0, 0],
                 [3, 2],
                 [0, 0]],
                [[0, 0],
                 [0, 0],
                 [1, 1],
                 [0, 1]]
            ])

            >>> cube.margin(axis=0)
            np.array([
                [3, 2],
                [3, 4],
                [4, 3],
                [0, 1],
            ])
        """

        table = self._counts(weighted).raw_cube_array
        new_axis = self._adjust_axis(axis)
        index = tuple(
            None if i in new_axis else slice(None) for i, _ in enumerate(table.shape)
        )

        # Calculate denominator. Only include those H&S dimensions, across
        # which we DON'T sum. These H&S are needed because of the shape, when
        # dividing. Those across dims which are summed across MUST NOT be
        # included, because they would change the result.
        hs_dims = self._hs_dims_for_den(include_transforms_for_dims, axis)
        den = self._apply_subtotals(
            self._apply_missings(table, include_missing=include_missing), hs_dims
        )

        # Apply correct mask (based on the as_array shape)
        arr = self._as_array(
            include_transforms_for_dims=hs_dims, include_missing=include_missing
        )

        # ---prune array if pruning was requested---
        if prune:
            arr = self._prune_body(arr, transforms=hs_dims)

        arr = self._drop_mr_cat_dims(arr, fix_valids=include_missing)

        if isinstance(arr, np.ma.core.MaskedArray):
            # Inflate the reduced version of the array, to match the
            # non-reduced version, for the purposes of creating the correct
            # mask. Create additional dimension (with no elements) where MR_CAT
            # dimensions should be. Don't inflate 0th dimension if it has only
            # a single element, because it's not being reduced
            # in self._drop_mr_cat_dims
            inflate_ind = tuple(
                (
                    None
                    if (
                        d.dimension_type == DT.MR_CAT
                        or i != 0
                        and (n <= 1 or len(d.valid_elements) <= 1)
                    )
                    else slice(None)
                )
                for i, (d, n) in enumerate(zip(self._all_dimensions, table.shape))
            )
            mask = np.logical_or(np.zeros(den.shape, dtype=bool), arr.mask[inflate_ind])
            den = np.ma.masked_array(den, mask)

        if (
            self.ndim != 1
            or axis is None
            or axis == 0
            and len(self._all_dimensions) == 1
        ):
            # Special case for 1D cube wigh MR, for "Table" direction
            den = np.sum(den, axis=new_axis)[index]

        den = self._drop_mr_cat_dims(
            den, fix_valids=(include_missing or include_mr_cat)
        )
        if den.shape[0] == 1 and len(den.shape) > 1 and self.ndim < 3:
            den = den.reshape(den.shape[1:])
        return den

    @lazyproperty
    def missing(self):
        """Get missing count of a cube."""
        return self._measures.missing_count

    @lazyproperty
    def mr_dim_ind(self):
        """Return int, tuple of int, or None, representing MR indices.

        The return value represents the index of each multiple-response (MR)
        dimension in this cube. Return value is None if there are no MR
        dimensions, and int if there is one MR dimension, and a tuple of int
        when there are more than one. The index is the (zero-based) position
        of the MR dimensions in the _ApparentDimensions sequence returned by
        the :attr"`.dimensions` property.
        """
        # TODO: rename to `mr_dim_idxs` or better yet get rid of need for
        # this as it's really a cube internal characteristic.
        # TODO: Make this return a tuple in all cases, like (), (1,), or (0, 2).
        indices = tuple(
            idx
            for idx, d in enumerate(self.dimensions)
            if d.dimension_type == DT.MR_SUBVAR
        )
        if indices == ():
            return None
        if len(indices) == 1:
            return indices[0]
        return indices

    @lazyproperty
    def name(self):
        """Return the name of the cube.

        If the cube has 2 diensions, return the name of the second one. In case
        of a different number of dimensions, default to returning the name of
        the last one. In case of no dimensions, return the empty string.
        """
        if not self.dimensions:
            return None
        return self.dimensions[0].name

    @lazyproperty
    def ndim(self):
        """int count of dimensions for this cube."""
        return len(self.dimensions)

    def percentages(self, axis=None):
        """Get the percentages for crunch cube values.

        This function calculates the percentages for crunch cube values. The
        percentages are based on the values of the 'proportions'.

        Args
            axis (int): Base axis of percentages calculation. If no axis is
                        provided, calculations are done accros entire table.

        Returns
            (nparray): Calculated array of crunch cube percentages.

        Example 1:
            >>> cube = CrunchCube(fixt_cat_x_cat)
            np.array([
               [5, 2],
               [5, 3],
            ])

            >>> cube.percentages()
            np.array([
                [33.33333, 13.33333],
                [33.33333, 20.00000],
            ])

        Example 2:
            >>> cube = CrunchCube(fixt_cat_x_cat)
            np.array([
               [5, 2],
               [5, 3],
            ])

            >>> cube.percentages(axis=0)
            np.array([
                [50., 40.],
                [50., 60.],
            ])
        """
        return self.proportions(axis) * 100

    def population_counts(
        self,
        population_size,
        weighted=True,
        include_missing=False,
        include_transforms_for_dims=None,
        prune=False,
    ):
        """Return counts scaled in proportion to overall population.

        The return value is a numpy.ndarray object. Count values are scaled
        proportionally to approximate their value if the entire population
        had been sampled. This calculation is based on the estimated size of
        the population provided as *population size*. The remaining arguments
        have the same meaning as they do for the `.proportions()` method.

        Example::

            >>> cube = CrunchCube(fixt_cat_x_cat)
            >>> cube.as_array()
            np.array([
               [5, 2],
               [5, 3],
            ])
            >>> cube.population_counts(9000)
            np.array([
                [3000, 1200],
                [3000, 1800],
            ])
        """
        population_counts = [
            slice_.population_counts(
                population_size,
                weighted=weighted,
                include_missing=include_missing,
                include_transforms_for_dims=include_transforms_for_dims,
                prune=prune,
            )
            for slice_ in self.slices
        ]

        if len(population_counts) > 1:
            return np.array(population_counts)
        return population_counts[0]

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """
        return self._measures.population_fraction

    def proportions(
        self,
        axis=None,
        weighted=True,
        include_transforms_for_dims=None,
        include_mr_cat=False,
        prune=False,
    ):
        """Return percentage values for cube as `numpy.ndarray`.

        This function calculates the proportions across the selected axis
        of a crunch cube. For most variable types, it means the value divided
        by the margin value. For a multiple-response variable, the value is
        divided by the sum of selected and non-selected slices.

        *axis* (int): base axis of proportions calculation. If no axis is
        provided, calculations are done across the entire table.

        *weighted* (bool): Specifies weighted or non-weighted proportions.

        *include_transforms_for_dims* (list): Also include headings and
        subtotals transformations for the provided dimensions. If the
        dimensions have the transformations, they'll be included in the
        resulting numpy array. If the dimensions don't have the
        transformations, nothing will happen (the result will be the same as
        if the argument weren't provided).

        *include_transforms_for_dims* (list): Include headers and subtotals
        (H&S) across various dimensions. The dimensions are provided as list
        elements. For example: "include_transforms_for_dims=[0, 1]" instructs
        the CrunchCube to return H&S for both rows and columns (if it's a 2D
        cube).

        *include_mr_cat* (bool): Include MR categories.

        *prune* (bool): Instructs the CrunchCube to prune empty rows/cols.
        Emptiness is determined by the state of the margin (if it's either
        0 or nan at certain index). If it is, the corresponding row/col is
        not included in the result.

        Example 1::

            >>> cube = CrunchCube(fixt_cat_x_cat)
            np.array([
               [5, 2],
               [5, 3],
            ])

            >>> cube.proportions()
            np.array([
                [0.3333333, 0.1333333],
                [0.3333333, 0.2000000],
            ])

        Example 2::

            >>> cube = CrunchCube(fixt_cat_x_cat)
            np.array([
               [5, 2],
               [5, 3],
            ])

            >>> cube.proportions(axis=0)
            np.array([
                [0.5, 0.4],
                [0.5, 0.6],
            ])
        """

        # Calculate numerator from table (include all H&S dimensions).
        table = self._measure(weighted).raw_cube_array
        num = self._apply_subtotals(
            self._apply_missings(table), include_transforms_for_dims
        )

        proportions = num / self._denominator(
            weighted, include_transforms_for_dims, axis
        )
        if not include_mr_cat:
            proportions = self._drop_mr_cat_dims(proportions)

        # Apply correct mask (based on the as_array shape)
        arr = self.as_array(
            prune=prune, include_transforms_for_dims=include_transforms_for_dims
        )
        if isinstance(arr, np.ma.core.MaskedArray):
            proportions = np.ma.masked_array(proportions, arr.mask)

        return proportions

    def _denominator(self, weighted, include_transforms_for_dims, axis):
        """Calculate denominator for percentages.

        Only include those H&S dimensions, across which we DON'T sum. These H&S
        are needed because of the shape, when dividing. Those across dims
        which are summed across MUST NOT be included, because they would
        change the result."""

        table = self._measure(weighted).raw_cube_array
        new_axis = self._adjust_axis(axis)
        index = tuple(
            None if i in new_axis else slice(None) for i, _ in enumerate(table.shape)
        )
        hs_dims = self._hs_dims_for_den(include_transforms_for_dims, axis)
        den = self._apply_subtotals(self._apply_missings(table), hs_dims)
        return np.sum(den, axis=new_axis)[index]

    def pvals(self, weighted=True, prune=False, hs_dims=None):
        """Return ndarray with calculated p-vals.

        This function calculates statistically significant results for
        categorical contingency tables. The values are calculated for 2D tables
        only. For 3D cubes, the slices' results are stacked together and
        returned as an ndarray.

        :param weighted: Use weighted counts for zscores
        :param prune: Prune based on unweighted counts
        :param hs_dims: Include headers and subtotals (as NaN values)
        :returns: 2 or 3 Dimensional ndarray, representing the p-values for each
                  cell of the table-like representation of the crunch cube.
        """
        res = [s.pvals(weighted, prune, hs_dims) for s in self.slices]
        return np.array(res) if self.ndim == 3 else res[0]

    @lazyproperty
    def row_direction_axis(self):
        # when dealing with 1D MR cubes, axis for margin should be 0
        if self.ndim == 1 and self.has_mr:
            return 0
        elif self.ndim == 3:
            return 2
        return 1

    def scale_means(self, hs_dims=None, prune=False):
        """Get cube means."""
        slices_means = [ScaleMeans(slice_).data for slice_ in self.slices]

        if hs_dims and self.ndim > 1:
            # Intersperse scale means with nans if H&S specified, and 2D. No
            # need to modify 1D, as only one mean will ever be inserted.
            inserted_indices = self.inserted_hs_indices()[-2:]
            for scale_means in slices_means:
                # Scale means 0 corresonds to the column dimension (is
                # calculated by using its values). The result of it, however,
                # is a row. That's why we need to check the insertions on the
                # row dim (inserted columns).
                if scale_means[0] is not None and 1 in hs_dims and inserted_indices[1]:
                    for i in inserted_indices[1]:
                        scale_means[0] = np.insert(scale_means[0], i, np.nan)
                # Scale means 1 is a column, so we need to check
                # for row insertions.
                if scale_means[1] is not None and 0 in hs_dims and inserted_indices[0]:
                    for i in inserted_indices[0]:
                        scale_means[1] = np.insert(scale_means[1], i, np.nan)

        if prune:
            # Apply pruning
            arr = self.as_array(include_transforms_for_dims=hs_dims, prune=True)
            if isinstance(arr, np.ma.core.MaskedArray):
                mask = arr.mask
                for i, scale_means in enumerate(slices_means):
                    if scale_means[0] is not None:
                        row_mask = (
                            mask.all(axis=0) if self.ndim < 3 else mask.all(axis=1)[i]
                        )
                        scale_means[0] = scale_means[0][~row_mask]
                    if self.ndim > 1 and scale_means[1] is not None:
                        col_mask = (
                            mask.all(axis=1) if self.ndim < 3 else mask.all(axis=2)[i]
                        )
                        scale_means[1] = scale_means[1][~col_mask]
        return slices_means

    @lazyproperty
    def slices(self):
        return self.get_slices()

    @lazyproperty
    def univariate_ca_main_axis(self):
        """For univariate CA, the main axis is the categorical axis"""
        return self.dim_types.index(DT.CA_CAT)

    def zscore(self, weighted=True, prune=False, hs_dims=None):
        """Return ndarray with cube's zscore measurements.

        Zscore is a measure of statistical significance of observed vs.
        expected counts. It's only applicable to a 2D contingency tables.
        For 3D cubes, the measures of separate slices are stacked together
        and returned as the result.

        :param weighted: Use weighted counts for zscores
        :param prune: Prune based on unweighted counts
        :param hs_dims: Include headers and subtotals (as NaN values)
        :returns zscore: ndarray representing zscore measurements
        """
        res = [s.zscore(weighted, prune, hs_dims) for s in self.slices]
        return np.array(res) if self.ndim == 3 else res[0]

    def wishart_pairwise_pvals(self, axis=0):
        """Return matrices of column-comparison p-values as list of numpy.ndarrays.

        Square, symmetric matrix along *axis* of pairwise p-values for the
        null hypothesis that col[i] = col[j] for each pair of columns.

        *axis* (int): axis along which to perform comparison. Only columns (0)
        are implemented currently.
        """
        return [slice_.wishart_pairwise_pvals(axis=axis) for slice_ in self.slices]

    def compare_to_column(self, slice_idx=0, column_idx=0):
        """Return matrices of column-comparison p-values as list of numpy.ndarrays.

        *t*-statistic and associated p-values for one column of one slice
        compared to each of the other columns in that slice.
        See scipy.stats.ttest_ind_from_stats.

        (It does not make sense to compare the same column at once across slices.)

        *slice_idx* (int): Index of slice for which we perform comparisons
        *column_idx* (int): Index of column to compare to, by default 0 (the first)
        """
        return self.slices[slice_idx].pairwise_significance_tests(column_idx)

    def _adjust_axis(self, axis):
        """Return raw axis/axes corresponding to apparent axis/axes.

        This method adjusts user provided 'axis' parameter, for some of the
        cube operations, mainly 'margin'. The user never sees the MR selections
        dimension, and treats all MRs as single dimensions. Thus we need to
        adjust the values of axis (to sum across) to what the user would've
        specified if he were aware of the existence of the MR selections
        dimension. The reason for this adjustment is that all of the operations
        performed troughout the margin calculations will be carried on an
        internal array, containing all the data (together with all selections).

        For more info on how it needs to operate, check the unit tests.
        """
        if not self._is_axis_allowed(axis):
            ca_error_msg = "Direction {} not allowed (items dimension)"
            raise ValueError(ca_error_msg.format(axis))

        if isinstance(axis, int):
            # If single axis was provided, create a list out of it, so that
            # we can do the subsequent iteration.
            axis = list([axis])
        elif axis is None:
            # If axis was None, create what user would expect in terms of
            # finding out the Total(s). In case of 2D cube, this will be the
            # axis of all the dimensions that the user can see, that is (0, 1),
            # because the selections dimension is invisible to the user. In
            # case of 3D cube, this will be the "total" across each slice, so
            # we need to drop the 0th dimension, and only take last two (1, 2).
            axis = range(self.ndim)[-2:]
        else:
            # In case of a tuple, just keep it as a list.
            axis = list(axis)
        axis = np.array(axis)

        # Create new array for storing updated values of axis. It's necessary
        # because it's hard to update the values in place.
        new_axis = np.array(axis)

        # Iterate over user-visible dimensions, and update axis when MR is
        # detected. For each detected MR, we need to increment all subsequent
        # axis (that were provided by the user). But we don't need to update
        # the axis that are "behind" the current MR.
        for i, dim in enumerate(self.dimensions):
            if dim.dimension_type == DT.MR_SUBVAR:
                # This formula updates only the axis that come "after" the
                # current MR (items) dimension.
                new_axis[axis >= i] += 1

        return tuple(new_axis)

    @staticmethod
    def _adjust_inserted_indices(inserted_indices_list, prune_indices_list):
        """Adjust inserted indices, if there are pruned elements."""
        # Created a copy, to preserve cached property
        updated_inserted = [[i for i in dim_inds] for dim_inds in inserted_indices_list]
        pruned_and_inserted = zip(prune_indices_list, updated_inserted)
        for prune_inds, inserted_inds in pruned_and_inserted:
            # Only prune indices if they're not H&S (inserted)
            prune_inds = prune_inds[~np.in1d(prune_inds, inserted_inds)]
            for i, ind in enumerate(inserted_inds):
                ind -= np.sum(prune_inds < ind)
                inserted_inds[i] = ind
        return updated_inserted

    @lazyproperty
    def _all_dimensions(self):
        """The AllDimensions object for this cube.

        The AllDimensions object provides access to all the dimensions
        appearing in the cube response, not only apparent dimensions (those
        that appear to a user). It also provides access to
        an _ApparentDimensions object which contains only those user-apparent
        dimensions (basically the categories dimension of each MR
        dimension-pair is suppressed).
        """
        return AllDimensions(dimension_dicts=self._cube_dict["result"]["dimensions"])

    def _apply_missings(self, res, include_missing=False):
        """Return ndarray with missing and insertions as specified.

        The return value is the result of the following operations on *res*,
        which is a raw cube value array (raw meaning it has shape of original
        cube response).

        * Remove vectors (rows/cols) for missing elements if *include_missin*
          is False.


        Note that it does *not* include pruning.
        """
        # --element idxs that satisfy `include_missing` arg. Note this
        # --includes MR_CAT elements so is essentially all-or-valid-elements
        element_idxs = tuple(
            (
                d.all_elements.element_idxs
                if include_missing
                else d.valid_elements.element_idxs
            )
            for d in self._all_dimensions
        )
        return res[np.ix_(*element_idxs)] if element_idxs else res

    def _apply_subtotals(self, res, include_transforms_for_dims):
        """* Insert subtotals (and perhaps other insertions later) for
          dimensions having their apparent dimension-idx in
          *include_transforms_for_dims*.
        """
        if not include_transforms_for_dims:
            return res

        suppressed_dim_count = 0
        for (dim_idx, dim) in enumerate(self._all_dimensions):
            if dim.dimension_type == DT.MR_CAT:
                suppressed_dim_count += 1
            # ---only marginable dimensions can be subtotaled---
            if not dim.is_marginable:
                continue
            apparent_dim_idx = dim_idx - suppressed_dim_count
            transform = (
                dim.has_transforms and apparent_dim_idx in include_transforms_for_dims
            )
            if not transform:
                continue
            # ---insert subtotals into result array---
            insertions = self._insertions(res, dim, dim_idx)
            res = self._update_result(res, insertions, dim_idx)

        return res

    def _as_array(
        self,
        include_missing=False,
        get_non_selected=False,
        weighted=True,
        include_transforms_for_dims=False,
    ):
        """Get crunch cube as ndarray.

        Args
            include_missing (bool): Include rows/cols for missing values.
            get_non_selected (bool): Get non-selected slices for MR vars.
            weighted (bool): Take weighted or unweighted counts.
            include_transforms_for_dims (list): For which dims to
                include headings & subtotals (H&S) transformations.
        Returns
            res (ndarray): Tabular representation of crunch cube
        """
        return self._apply_subtotals(
            self._apply_missings(
                self._measure(weighted).raw_cube_array, include_missing=include_missing
            ),
            include_transforms_for_dims,
        )

    @classmethod
    def _calculate_constraints_sum(cls, prop_table, prop_margin, axis):
        """Calculate sum of constraints (part of the standard error equation).

        This method calculates the sum of the cell proportions multiplied by
        row (or column) marginal proportions (margins divide by the total
        count). It does this by utilizing the matrix multiplication, which
        directly translates to the mathematical definition (the sum
        across i and j indices).
        """
        if axis not in [0, 1]:
            raise ValueError("Unexpected value for `axis`: {}".format(axis))

        V = prop_table * (1 - prop_table)
        if axis == 0:
            # If axis is 0, sumation is performed across the 'i' index, which
            # requires the matrix to be multiplied from the right
            # (because of the inner matrix dimensions).
            return np.dot(V, prop_margin)
        elif axis == 1:
            # If axis is 1, sumation is performed across the 'j' index, which
            # requires the matrix to be multiplied from the left
            # (because of the inner matrix dimensions).
            return np.dot(prop_margin, V)

    @lazyproperty
    def _col_direction_axis(self):
        return self.ndim - 2

    def _counts(self, weighted):
        """Return _BaseMeasure subclass for *weighted* counts.

        The return value is a _WeightedCountMeasure object if *weighted* is
        True and the cube response is weighted. Otherwise it is an
        _UnweightedCountMeasure object. Any means measure that may be present
        is not considered. Contrast with `._measure()` below.
        """
        return (
            self._measures.weighted_counts
            if weighted
            else self._measures.unweighted_counts
        )

    @staticmethod
    def _create_mask(res, row_prune_inds, col_prune_inds):
        mask_rows = np.repeat(row_prune_inds[:, None], len(col_prune_inds), axis=1)
        mask_cols = np.repeat(col_prune_inds[None, :], len(row_prune_inds), axis=0)
        return np.logical_or(mask_rows, mask_cols)

    @lazyproperty
    def _cube_dict(self):
        """dict containing raw cube response, parsed from JSON payload."""
        try:
            cube_response = self._cube_response_arg
            # ---parse JSON to a dict when constructed with JSON---
            cube_dict = (
                cube_response
                if isinstance(cube_response, dict)
                else json.loads(cube_response)
            )
            # ---cube is 'value' item in a shoji response---
            return cube_dict.get("value", cube_dict)
        except TypeError:
            raise TypeError(
                "Unsupported type <%s> provided. Cube response must be JSON "
                "(str) or dict." % type(self._cube_response_arg).__name__
            )

    def _drop_mr_cat_dims(self, array, fix_valids=False):
        """Return ndarray reflecting *array* with MR_CAT dims dropped.

        If any (except 1st) dimension has a single element, it is
        flattened in the resulting array (which is more convenient for the
        users of the CrunchCube).

        If the original shape of the cube is needed (e.g. to calculate the
        margins with correct axis arguments), this needs to happen before the
        call to this method '_drop_mr_cat_dims'.
        """
        # TODO: We cannot arbitrarily drop any dimension simply because it
        # has a length (shape) of 1. We must target MR_CAT dimensions
        # specifically. Otherwise unexpected results can occur based on
        # accidents of cube category count etc. If "user-friendly" reshaping
        # needs be done, it should be as a very last step and much safer to
        # leave that to the cr.cube client; software being "helpful" almost
        # never is.

        if not array.shape or len(array.shape) != len(self._all_dimensions):
            # This condition covers two cases:
            # 1. In case of no dimensions, the shape of the array is empty
            # 2. If the shape was already fixed, we don't need to fix it again.
            # This might happen while constructing the masked arrays. In case
            # of MR, we will have the selections dimension included thoughout
            # the calculations, and will only remove it before returning the
            # result to the user.
            return array

        # We keep MR selections (MR_CAT) dimensions in the array, all the way
        # up to here. At this point, we need to remove the non-selected part of
        # selections dimension (and subsequently purge the dimension itself).

        display_ind = (
            tuple(
                0 if dim.dimension_type == DT.MR_CAT else slice(None)
                for dim, n in zip(self._all_dimensions, array.shape)
            )
            if not fix_valids
            else np.ix_(
                *[
                    dim.valid_elements.element_idxs if n > 1 else [0]
                    for dim, n in zip(self._all_dimensions, array.shape)
                ]
            )
        )
        array = array[display_ind]

        # If a first dimension only has one element, we don't want to
        # remove it from the shape. Hence the i == 0 part. For other dimensions
        # that have one element, it means that these are the remnants of the MR
        # selections, which we don't need as separate dimensions.
        new_shape = [
            length for (i, length) in enumerate(array.shape) if length != 1 or i == 0
        ]
        return array.reshape(new_shape)

    @classmethod
    def _fix_valid_indices(cls, valid_indices, insertion_index, dim):
        """Add indices for H&S inserted elements."""
        # TODO: make this accept an immutable sequence for valid_indices
        # (a tuple) and return an immutable sequence rather than mutating an
        # argument.
        indices = np.array(sorted(valid_indices[dim]))
        slice_index = np.sum(indices <= insertion_index)
        indices[slice_index:] += 1
        indices = np.insert(indices, slice_index, insertion_index + 1)
        valid_indices[dim] = indices.tolist()
        return valid_indices

    @staticmethod
    def _hs_dims_for_den(hs_dims, axis):
        if axis is None or hs_dims is None:
            return None
        if isinstance(axis, int):
            axis = [axis]
        return [dim for dim in hs_dims if dim not in axis]

    def _inserted_dim_inds(self, transform_dims, axis):
        dim_ind = axis if self.ndim < 3 else axis + 1
        if not transform_dims or dim_ind not in transform_dims:
            return np.array([])

        inserted_inds = self.inserted_hs_indices()
        return np.array(inserted_inds[dim_ind] if len(inserted_inds) else [])

    def _insertions(self, result, dimension, dimension_index):
        """Return list of (idx, sum) pairs representing subtotals.

        *idx* is the int offset at which to insert the ndarray subtotal
        in *sum*.
        """

        def iter_insertions():
            for anchor_idx, addend_idxs in dimension.hs_indices:
                insertion_idx = (
                    -1
                    if anchor_idx == "top"
                    else result.shape[dimension_index] - 1
                    if anchor_idx == "bottom"
                    else anchor_idx
                )
                addend_fancy_idx = tuple(
                    [slice(None) for _ in range(dimension_index)]
                    + [np.array(addend_idxs)]
                )
                yield (
                    insertion_idx,
                    np.sum(result[addend_fancy_idx], axis=dimension_index),
                )

        return [insertion for insertion in iter_insertions()]

    def _is_axis_allowed(self, axis):
        """Check if axis are allowed.

        In case the calculation is requested over CA items dimension, it is not
        valid. It's valid in all other cases.
        """
        if axis is None:
            # If table direction was requested, we must ensure that each slice
            # doesn't have the CA items dimension (thus the [-2:] part). It's
            # OK for the 0th dimension to be items, since no calculation is
            # performed over it.
            if DT.CA_SUBVAR in self.dim_types[-2:]:
                return False
            return True

        if isinstance(axis, int):
            if self.ndim == 1 and axis == 1:
                # Special allowed case of a 1D cube, where "row"
                # directions is requested.
                return True
            axis = [axis]

        # ---axis is a tuple---
        for dim_idx in axis:
            if self.dim_types[dim_idx] == DT.CA_SUBVAR:
                # If any of the directions explicitly asked for directly
                # corresponds to the CA items dimension, the requested
                # calculation is not valid.
                return False

        return True

    @staticmethod
    def _margin_pruned_indices(margin, inserted_ind, axis):
        pruned_ind = np.logical_or(margin == 0, np.isnan(margin))

        if pruned_ind.ndim == 0:
            pruned_ind = np.array([pruned_ind])

        if np.any(margin) and inserted_ind is not None and any(inserted_ind):
            ind_inserted = np.zeros(pruned_ind.shape, dtype=bool)
            if len(pruned_ind.shape) == 2 and axis == 1:
                ind_inserted[:, inserted_ind] = True
            else:
                ind_inserted[inserted_ind] = True
            pruned_ind = np.logical_and(pruned_ind, ~ind_inserted)

        return pruned_ind

    def _measure(self, weighted):
        """_BaseMeasure subclass representing primary measure for this cube.

        If the cube response includes a means measure, the return value is
        means. Otherwise it is counts, with the choice between weighted or
        unweighted determined by *weighted*.

        Note that weighted counts are provided on an "as-available" basis.
        When *weighted* is True and the cube response is not weighted,
        unweighted counts are returned.
        """
        return (
            self._measures.means
            if self._measures.means is not None
            else self._measures.weighted_counts
            if weighted
            else self._measures.unweighted_counts
        )

    @lazyproperty
    def _measures(self):
        """_Measures object for this cube.

        Provides access to unweighted counts, and weighted counts and/or means
        when available.
        """
        return _Measures(self._cube_dict, self._all_dimensions)

    def _prune_3d_body(self, res, transforms):
        """Return masked array where mask indicates pruned vectors.

        *res* is an ndarray (result). *transforms* is a list of ...
        """
        mask = np.zeros(res.shape)
        mr_dim_idxs = self.mr_dim_ind

        for i, prune_inds in enumerate(self.prune_indices(transforms)):
            rows_pruned = prune_inds[0]
            cols_pruned = prune_inds[1]
            rows_pruned = np.repeat(rows_pruned[:, None], len(cols_pruned), axis=1)
            cols_pruned = np.repeat(cols_pruned[None, :], len(rows_pruned), axis=0)
            slice_mask = np.logical_or(rows_pruned, cols_pruned)

            # In case of MRs we need to "inflate" mask
            if mr_dim_idxs == (1, 2):
                slice_mask = slice_mask[:, np.newaxis, :, np.newaxis]
            elif mr_dim_idxs == (0, 1):
                slice_mask = slice_mask[np.newaxis, :, np.newaxis, :]
            elif mr_dim_idxs == (0, 2):
                slice_mask = slice_mask[np.newaxis, :, :, np.newaxis]
            elif mr_dim_idxs == 1 and self.ndim == 3:
                slice_mask = slice_mask[:, np.newaxis, :]
            elif mr_dim_idxs == 2 and self.ndim == 3:
                slice_mask = slice_mask[:, :, np.newaxis]

            mask[i] = slice_mask

        res = np.ma.masked_array(res, mask=mask)
        return res

    def _prune_3d_indices(self, transforms):
        row_margin = self._pruning_base(
            hs_dims=transforms, axis=self.row_direction_axis
        )
        col_margin = self._pruning_base(
            hs_dims=transforms, axis=self._col_direction_axis
        )
        return [
            self._prune_indices_tuple(rm, cm, transforms)
            for rm, cm in zip(row_margin, col_margin)
        ]

    def _prune_body(self, res, transforms=None):
        """Return a masked version of *res* where pruned rows/cols are masked.

        Return value is an `np.ma.MaskedArray` object. Pruning is the removal
        of rows or columns whose corresponding marginal elements are either
        0 or not defined (np.nan).
        """
        if self.ndim > 2:
            return self._prune_3d_body(res, transforms)

        res = self._drop_mr_cat_dims(res)

        # ---determine which rows should be pruned---
        row_margin = self._pruning_base(
            hs_dims=transforms, axis=self.row_direction_axis
        )
        # ---adjust special-case row-margin values---
        item_types = (DT.MR, DT.CA_SUBVAR)
        if self.ndim > 1 and self.dim_types[1] in item_types and len(res.shape) > 1:
            # ---when row-dimension has only one category it gets squashed---
            axis = 1 if res.shape[0] > 1 else None
            # ---in CAT x MR case (or if it has CA subvars) we get
            # a 2D margin (denom really)---
            row_margin = np.sum(row_margin, axis=axis)
        row_prune_inds = self._margin_pruned_indices(
            row_margin, self._inserted_dim_inds(transforms, 0), 0
        )

        # ---a 1D only has rows, so mask only with row-prune-idxs---
        if self.ndim == 1 or len(res.shape) == 1:
            # For 1D, margin is calculated as the row margin.
            return np.ma.masked_array(res, mask=row_prune_inds)

        # ---determine which columns should be pruned---
        col_margin = self._pruning_base(
            hs_dims=transforms, axis=self._col_direction_axis
        )
        if col_margin.ndim > 1:
            # In case of MR x CAT, we have 2D margin
            col_margin = np.sum(col_margin, axis=0)
        col_prune_inds = self._margin_pruned_indices(
            col_margin, self._inserted_dim_inds(transforms, 1), 1
        )

        # ---create rows x cols mask and mask the result array---
        mask = self._create_mask(res, row_prune_inds, col_prune_inds)
        res = np.ma.masked_array(res, mask=mask)

        # ---return the masked array---
        return res

    def prune_indices(self, transforms=None):
        """Return indices of pruned rows and columns as list.

        The return value has one of three possible forms:

        * a 1-element list of row indices (in case of 1D cube)
        * 2-element list of row and col indices (in case of 2D cube)
        * n-element list of tuples of 2 elements (if it's 3D cube).

        For each case, the 2 elements are the ROW and COL indices of the
        elements that need to be pruned. If it's a 3D cube, these indices are
        calculated "per slice", that is NOT on the 0th dimension (as the 0th
        dimension represents the slices).
        """
        if self.ndim >= 3:
            # In case of a 3D cube, return list of tuples
            # (of row and col pruned indices).
            return self._prune_3d_indices(transforms)

        def prune_non_3d_indices(transforms):
            row_margin = self._pruning_base(
                hs_dims=transforms, axis=self.row_direction_axis
            )
            row_indices = self._margin_pruned_indices(
                row_margin, self._inserted_dim_inds(transforms, 0), 0
            )

            if row_indices.ndim > 1:
                # In case of MR, we'd have 2D prune indices
                row_indices = row_indices.all(axis=1)

            if self.ndim == 1:
                return [row_indices]

            col_margin = self._pruning_base(
                hs_dims=transforms, axis=self._col_direction_axis
            )

            col_indices = self._margin_pruned_indices(
                col_margin, self._inserted_dim_inds(transforms, 1), 1
            )
            if col_indices.ndim > 1:
                # In case of MR, we'd have 2D prune indices
                col_indices = col_indices.all(axis=0)

            return [row_indices, col_indices]

        # In case of 1 or 2 D cubes, return a list of
        # row indices (or row and col indices)
        return prune_non_3d_indices(transforms)

    def _prune_indices_tuple(self, row_margin, column_margin, transforms):
        if row_margin.ndim > 1:
            row_margin = np.sum(row_margin, axis=1)
        if column_margin.ndim > 1:
            column_margin = np.sum(column_margin, axis=0)

        return (
            self._margin_pruned_indices(
                row_margin, self._inserted_dim_inds(transforms, 0), 0
            ),
            self._margin_pruned_indices(
                column_margin, self._inserted_dim_inds(transforms, 1), 1
            ),
        )

    def _pruning_base(self, axis=None, hs_dims=None):
        """Gets margin if across CAT dimension. Gets counts if across items.

        Categorical variables are pruned based on their marginal values. If the
        marginal is a 0 or a NaN, the corresponding row/column is pruned. In
        case of a subvars (items) dimension, we only prune if all the counts
        of the corresponding row/column are zero.
        """
        if not self._is_axis_allowed(axis):
            # In case we encountered axis that would go across items dimension,
            # we need to return at least some result, to prevent explicitly
            # checking for this condition, wherever self._margin is used
            return self.as_array(weighted=False, include_transforms_for_dims=hs_dims)

        # In case of allowed axis, just return the normal API margin. This call
        # would throw an exception when directly invoked with bad axis. This is
        # intended, because we want to be as explicit as possible. Margins
        # across items are not allowed.
        return self.margin(
            axis=axis, weighted=False, include_transforms_for_dims=hs_dims
        )

    def _update_result(self, result, insertions, dimension_index):
        """Insert subtotals into resulting ndarray."""
        for j, (ind_insertion, value) in enumerate(insertions):
            result = np.insert(
                result, ind_insertion + j + 1, value, axis=dimension_index
            )
        return result


class _Measures(object):
    """Provides access to measures contained in cube response."""

    def __init__(self, cube_dict, all_dimensions):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions

    @lazyproperty
    def is_weighted(self):
        """True if weights have been applied to the measure(s) for this cube.

        Unweighted counts are available for all cubes. Weighting applies to
        any other measures provided by the cube.
        """
        cube_dict = self._cube_dict
        if cube_dict.get("query", {}).get("weight") is not None:
            return True
        if cube_dict.get("weight_var") is not None:
            return True
        if cube_dict.get("weight_url") is not None:
            return True
        unweighted_counts = cube_dict["result"]["counts"]
        count_data = cube_dict["result"]["measures"].get("count", {}).get("data")
        if unweighted_counts != count_data:
            return True
        return False

    @lazyproperty
    def means(self):
        """_MeanMeasure object providing access to means values.

        None when the cube response does not contain a mean measure.
        """
        mean_measure_dict = (
            self._cube_dict.get("result", {}).get("measures", {}).get("mean")
        )
        if mean_measure_dict is None:
            return None
        return _MeanMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows in cube response."""
        if self.means:
            return self.means.missing_count
        return self._cube_dict["result"].get("missing", 0)

    @lazyproperty
    def population_fraction(self):
        """The filtered/unfiltered ratio for cube response.

        This value is required for properly calculating population on a cube
        where a filter has been applied. Returns 1.0 for an unfiltered cube.
        Returns `np.nan` if the unfiltered count is zero, which would
        otherwise result in a divide-by-zero error.
        """
        numerator = self._cube_dict["result"].get("filtered", {}).get("weighted_n")
        denominator = self._cube_dict["result"].get("unfiltered", {}).get("weighted_n")
        try:
            return numerator / denominator
        except ZeroDivisionError:
            return np.nan
        except Exception:
            return 1.0

    @lazyproperty
    def unweighted_counts(self):
        """_UnweightedCountMeasure object for this cube.

        This object provides access to unweighted counts for this cube,
        whether or not the cube contains weighted counts.
        """
        return _UnweightedCountMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def unweighted_n(self):
        """int count of actual rows represented by query response."""
        return self._cube_dict["result"]["n"]

    @lazyproperty
    def weighted_counts(self):
        """_WeightedCountMeasure object for this cube.

        This object provides access to weighted counts for this cube, if
        available. If the cube response is not weighted, the
        _UnweightedCountMeasure object for this cube is returned.
        """
        if not self.is_weighted:
            return self.unweighted_counts
        return _WeightedCountMeasure(self._cube_dict, self._all_dimensions)

    @lazyproperty
    def weighted_n(self):
        """float count of returned rows adjusted for weighting."""
        if not self.is_weighted:
            return float(self.unweighted_n)
        return float(sum(self._cube_dict["result"]["measures"]["count"]["data"]))


class _BaseMeasure(object):
    """Base class for measure objects."""

    def __init__(self, cube_dict, all_dimensions):
        self._cube_dict = cube_dict
        self._all_dimensions = all_dimensions

    @lazyproperty
    def raw_cube_array(self):
        """Return read-only ndarray of measure values from cube-response.

        The shape of the ndarray mirrors the shape of the (raw) cube
        response. Specifically, it includes values for missing elements, any
        MR_CAT dimensions, and any prunable rows and columns.
        """
        array = np.array(self._flat_values).reshape(self._all_dimensions.shape)
        # ---must be read-only to avoid hard-to-find bugs---
        array.flags.writeable = False
        return array

    @lazyproperty
    def _flat_values(self):
        """Return tuple of mean values as found in cube response.

        This property must be implemented by each subclass.
        """
        raise NotImplementedError("must be implemented by each subclass")


class _MeanMeasure(_BaseMeasure):
    """Statistical mean values from a cube-response."""

    @lazyproperty
    def missing_count(self):
        """numeric representing count of missing rows reflected in response."""
        return self._cube_dict["result"]["measures"]["mean"].get("n_missing", 0)

    @lazyproperty
    def _flat_values(self):
        """Return tuple of mean values as found in cube response.

        Mean data may include missing items represented by a dict like
        {'?': -1} in the cube response. These are replaced by np.nan in the
        returned value.
        """
        return tuple(
            np.nan if type(x) is dict else x
            for x in self._cube_dict["result"]["measures"]["mean"]["data"]
        )


class _UnweightedCountMeasure(_BaseMeasure):
    """Unweighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """tuple of int counts before weighting."""
        return tuple(self._cube_dict["result"]["counts"])


class _WeightedCountMeasure(_BaseMeasure):
    """Weighted counts for cube."""

    @lazyproperty
    def _flat_values(self):
        """tuple of numeric counts after weighting."""
        return tuple(self._cube_dict["result"]["measures"]["count"]["data"])
